"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Optional, Sequence
import torch.nn as nn


class ClassifierFCN(nn.Module):
    """
    Sequence of (Conv2D 1X1 , ReLU, Dropout). The length of the sequence and layers size defined by layers_description
    """

    def __init__(
        self,
        in_ch: int,
        num_classes: Optional[int],
        layers_description: Sequence[int] = (256,),
        dropout_rate: float = 0.1,
    ):
        """
        :param in_ch: Number of input channels
        :param num_classes: Appends Conv2D(last_layer_size, num_classes, kernel_size=1, stride=1) if num_classes is not None
        :param layers_description: defines the length of the sequence and layers size.
        :param dropout_rate: if 0 will not include the dropout layers
        """
        super().__init__()
        layer_list = []
        layer_list.append(nn.Conv2d(in_ch, layers_description[0], kernel_size=1, stride=1))
        layer_list.append(nn.ReLU())
        if dropout_rate is not None and dropout_rate > 0:
            layer_list.append(nn.Dropout(p=dropout_rate))
        last_layer_size = layers_description[0]
        for curr_layer_size in layers_description[1:]:
            layer_list.append(nn.Conv2d(last_layer_size, curr_layer_size, kernel_size=1, stride=1))
            layer_list.append(nn.ReLU())
            if dropout_rate is not None and dropout_rate > 0:
                layer_list.append(nn.Dropout(p=dropout_rate))
            last_layer_size = curr_layer_size

        if num_classes is not None:
            layer_list.append(nn.Conv2d(last_layer_size, num_classes, kernel_size=1, stride=1))

        self.classifier = nn.Sequential(*layer_list)

    def forward(self, x):
        x = self.classifier(x)
        return x


class ClassifierFCN3D(nn.Module):
    """
    Sequence of (Conv3D 1X1 , ReLU, Dropout). The length of the sequence and layers size defined by layers_description
    """

    def __init__(
        self,
        in_ch: int,
        num_classes: Optional[int],
        layers_description: Sequence[int] = (256,),
        dropout_rate: float = 0.1,
    ):
        """
        :param in_ch: Number of input channels
        :param num_classes: Appends Conv2D(last_layer_size, num_classes, kernel_size=1, stride=1) if num_classes is not None
        :param layers_description: defines the length of the sequence and layers size.
        :param dropout_rate: if 0 will not include the dropout layers
        """
        super().__init__()
        layer_list = []
        layer_list.append(nn.Conv3d(in_ch, layers_description[0], kernel_size=1, stride=1))
        layer_list.append(nn.ReLU())
        if dropout_rate is not None and dropout_rate > 0:
            layer_list.append(nn.Dropout(p=dropout_rate))
        last_layer_size = layers_description[0]
        for curr_layer_size in layers_description[1:]:
            layer_list.append(nn.Conv3d(last_layer_size, curr_layer_size, kernel_size=1, stride=1))
            layer_list.append(nn.ReLU())
            if dropout_rate is not None and dropout_rate > 0:
                layer_list.append(nn.Dropout(p=dropout_rate))
            last_layer_size = curr_layer_size

        if num_classes is not None:
            layer_list.append(nn.Conv3d(last_layer_size, num_classes, kernel_size=1, stride=1))

        self.classifier = nn.Sequential(*layer_list)

    def forward(self, x):
        x = self.classifier(x)
        return x


class ClassifierMLP(nn.Module):
    """
    Sequence of (Linear , ReLU, Dropout). The length of the sequence and layers size defined by layers_description
    """

    def __init__(
        self,
        in_ch: int,
        num_classes: Optional[int],
        layers_description: Sequence[int] = (256,),
        dropout_rate: float = 0.1,
    ):
        """
        :param in_ch: Number of input channels
        :param num_classes: Appends Conv2D(last_layer_size, num_classes, kernel_size=1, stride=1) if num_classes is not None
        :param layers_description: defines the length of the sequence and layers size.
        :param dropout_rate: if 0 will not include the dropout layers
        """
        super().__init__()
        layer_list = []
        layer_list.append(nn.Linear(in_ch, layers_description[0]))
        layer_list.append(nn.ReLU())
        if dropout_rate is not None and dropout_rate > 0:
            layer_list.append(nn.Dropout(p=dropout_rate))
        last_layer_size = layers_description[0]
        for curr_layer_size in layers_description[1:]:
            layer_list.append(nn.Linear(last_layer_size, curr_layer_size))
            layer_list.append(nn.ReLU())
            if dropout_rate is not None and dropout_rate > 0:
                layer_list.append(nn.Dropout(p=dropout_rate))
            last_layer_size = curr_layer_size

        if num_classes is not None:
            layer_list.append(nn.Linear(last_layer_size, num_classes))

        self.classifier = nn.Sequential(*layer_list)

    def forward(self, x):
        x = self.classifier(x)
        return x



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Callable, Optional

import torch
from torch.hub import load_state_dict_from_url
import torch.nn as nn

from urllib.error import URLError
import logging


def make_seq(foo: Callable, num: int, *args, **kwargs):
    """
    Makes a sequence of blocks
    """
    l = [foo(*args, **kwargs) for i in range(num)]
    return nn.Sequential(*l)


def make_final_seq(foo: Callable, num: int, *args, **kwargs):
    """
    Makes a sequence of blocks, but passes 'final_block'= True (to cut inside final block17)
    """
    l = [foo(*args, **kwargs) for i in range(num - 1)]
    kwargs["final_block"] = True
    l.append(foo(*args, **kwargs))
    return nn.Sequential(*l)


class BasicConv2d(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride, padding=0):
        super(BasicConv2d, self).__init__()
        self.conv = nn.Conv2d(
            in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=padding, bias=False
        )  # verify bias false
        self.bn = nn.BatchNorm2d(out_planes, eps=0.001, momentum=0.01, affine=True)  # changed from original 0.1
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Mixed_5b(nn.Module):
    def __init__(self):
        super(Mixed_5b, self).__init__()

        self.branch0 = BasicConv2d(192, 96, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(192, 48, kernel_size=1, stride=1), BasicConv2d(48, 64, kernel_size=5, stride=1, padding=2)
        )

        self.branch2 = nn.Sequential(
            BasicConv2d(192, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 96, kernel_size=3, stride=1, padding=1),
            BasicConv2d(96, 96, kernel_size=3, stride=1, padding=1),
        )

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False), BasicConv2d(192, 64, kernel_size=1, stride=1)
        )

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out  # results in 96+64+96+64 channels


class Block35(nn.Module):
    def __init__(self, scale=1.0):
        super(Block35, self).__init__()

        self.scale = scale

        self.branch0 = BasicConv2d(320, 32, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(320, 32, kernel_size=1, stride=1), BasicConv2d(32, 32, kernel_size=3, stride=1, padding=1)
        )

        self.branch2 = nn.Sequential(
            BasicConv2d(320, 32, kernel_size=1, stride=1),
            BasicConv2d(32, 48, kernel_size=3, stride=1, padding=1),
            BasicConv2d(48, 64, kernel_size=3, stride=1, padding=1),
        )

        self.conv2d = nn.Conv2d(128, 320, kernel_size=1, stride=1)
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        out = self.relu(out)
        return out  # results in 320 channels


class Mixed_6a(nn.Module):
    def __init__(self):
        super(Mixed_6a, self).__init__()

        self.branch0 = BasicConv2d(320, 384, kernel_size=3, stride=2)

        self.branch1 = nn.Sequential(
            BasicConv2d(320, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 256, kernel_size=3, stride=1, padding=1),
            BasicConv2d(256, 384, kernel_size=3, stride=2),
        )

        self.branch2 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        return out  # results in 384+384+320


class Block17(nn.Module):
    def __init__(self, scale=1.0, final_block=False, intra_block_cut_level=384):
        super(Block17, self).__init__()

        self.scale = scale
        self.final_block = final_block
        self.intra_block_cut_level = intra_block_cut_level

        self.branch0 = BasicConv2d(1088, 192, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(1088, 128, kernel_size=1, stride=1),
            BasicConv2d(128, 160, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(160, 192, kernel_size=(7, 1), stride=1, padding=(3, 0)),
        )

        self.conv2d = nn.Conv2d(384, 1088, kernel_size=1, stride=1)
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        if self.final_block and self.intra_block_cut_level == 384:
            return out  # results in 384 channels

        out = self.conv2d(out)
        out = out * self.scale + x
        if self.final_block and self.intra_block_cut_level == 1088:
            return out  # results in 1088 channels
        out = self.relu(out)
        return out  # results in 1088 channels


class Mixed_7a(nn.Module):
    def __init__(self):
        super(Mixed_7a, self).__init__()

        self.branch0 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1), BasicConv2d(256, 384, kernel_size=3, stride=2)
        )

        self.branch1 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1), BasicConv2d(256, 288, kernel_size=3, stride=2)
        )

        self.branch2 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 288, kernel_size=3, stride=1, padding=1),
            BasicConv2d(288, 320, kernel_size=3, stride=2),
        )

        self.branch3 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out  # results in 384+288+320+1088


class Block8(nn.Module):
    def __init__(self, scale=1.0, noReLU=False):
        super(Block8, self).__init__()

        self.scale = scale
        self.noReLU = noReLU

        self.branch0 = BasicConv2d(2080, 192, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(2080, 192, kernel_size=1, stride=1),
            BasicConv2d(192, 224, kernel_size=(1, 3), stride=1, padding=(0, 1)),
            BasicConv2d(224, 256, kernel_size=(3, 1), stride=1, padding=(1, 0)),
        )

        self.conv2d = nn.Conv2d(448, 2080, kernel_size=1, stride=1)
        if not self.noReLU:
            self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        if not self.noReLU:
            out = self.relu(out)
        return out  # results in 2080 channels


class BackboneInceptionResnetV2(nn.Module):
    def __init__(
        self,
        logical_units_num: int = 14,
        intra_block_cut_level: int = 384,
        input_channels_num: int = 3,
        pretrained_weights_url: Optional[
            str
        ] = "http://data.lip6.fr/cadene/pretrainedmodels/inceptionresnetv2-520b38e4.pth",
    ) -> None:

        super().__init__()
        self.logical_units_num = logical_units_num
        self.intra_block_cut_level = intra_block_cut_level
        self.input_channels_num = input_channels_num

        # Modules
        self.conv2d_1a = BasicConv2d(3, 32, kernel_size=3, stride=2)
        self.conv2d_2a = BasicConv2d(32, 32, kernel_size=3, stride=1)
        self.conv2d_2b = BasicConv2d(32, 64, kernel_size=3, stride=1, padding=1)
        self.maxpool_3a = nn.MaxPool2d(3, stride=2)
        self.conv2d_3b = BasicConv2d(64, 80, kernel_size=1, stride=1)
        self.conv2d_4a = BasicConv2d(80, 192, kernel_size=3, stride=1)
        self.maxpool_5a = nn.MaxPool2d(3, stride=2)
        self.mixed_5b = Mixed_5b()
        self.feature_depth = 96 + 64 + 96 + 64

        if self.logical_units_num >= 1:
            self.repeat = make_seq(Block35, min(self.logical_units_num - 1 + 1, 10), scale=0.17)
            self.feature_depth = 320

        if self.logical_units_num >= 11:
            self.mixed_6a = Mixed_6a()
            self.feature_depth = 384 + 384 + 320

        if self.logical_units_num >= 12:
            if 12 <= self.logical_units_num < 32:
                self.repeat_1 = make_final_seq(
                    Block17,
                    min(self.logical_units_num - 12 + 1, 20),
                    scale=0.10,
                    intra_block_cut_level=self.intra_block_cut_level,
                )
                self.feature_depth = self.intra_block_cut_level

            else:
                self.repeat_1 = make_seq(Block17, min(self.logical_units_num - 12 + 1, 20), scale=0.10)
                self.feature_depth = 1088

        if self.logical_units_num >= 32:
            self.mixed_7a = Mixed_7a()
            self.feature_depth = 384 + 288 + 320 + 1088

        if self.logical_units_num >= 33:
            self.repeat_2 = make_seq(Block8, min(self.logical_units_num - 33 + 1, 9), scale=0.20)
            self.feature_depth = 2080

        if self.logical_units_num >= 42:
            self.block8 = Block8(noReLU=True)
            self.feature_depth = 2080

        if self.logical_units_num >= 43:
            self.conv2d_7b = BasicConv2d(2080, 1536, kernel_size=1, stride=1)
            self.feature_depth = 1536

        # load pretrained parameters if required
        if pretrained_weights_url is not None:
            try:
                state_dict = load_state_dict_from_url(pretrained_weights_url)
                self.load_state_dict(state_dict, strict=False)
            except AttributeError:
                raise Exception("Invalid URL for InceptionResnetV2 pretrained weights")
            except URLError as e:
                logging.getLogger("Fuse").warning(
                    f"Couldn't load pretraind weights from the url: '{pretrained_weights_url}' due to the following URLError: '{e}'."
                )

        # recreate the first conv with the required number of input parameters
        if input_channels_num != 3:
            self.conv2d_1a = BasicConv2d(input_channels_num, 32, kernel_size=3, stride=2)

    def features(self, input_tensor):
        x = self.conv2d_1a(input_tensor)
        x = self.conv2d_2a(x)
        x = self.conv2d_2b(x)
        x = self.maxpool_3a(x)
        x = self.conv2d_3b(x)
        x = self.conv2d_4a(x)
        x = self.maxpool_5a(x)
        x = self.mixed_5b(x)

        if self.logical_units_num >= 1:
            x = self.repeat(x)

        if self.logical_units_num >= 11:
            x = self.mixed_6a(x)

        if self.logical_units_num >= 12:
            x = self.repeat_1(x)

        if self.logical_units_num >= 32:
            x = self.mixed_7a(x)

        if self.logical_units_num >= 33:
            x = self.repeat_2(x)

        if self.logical_units_num >= 42:
            x = self.block8(x)

        if self.logical_units_num >= 43:
            x = self.conv2d_7b(x)
        return x

    def forward(self, input_tensor):
        feature_map = self.features(
            input_tensor
        )  # typical features shape when cutting at level 14 is [batch_size, 384, H, W]
        return feature_map

    def get_feature_depth(self):
        return self.feature_depth



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from abc import abstractmethod
from typing import Dict, Hashable, List, Optional, Sequence, Union

from torch.utils.data.dataset import Dataset


class DatasetBase(Dataset):
    @abstractmethod
    def create(self, **kwargs) -> None:
        """
        Make the dataset operational: might include data caching, reloading and more.
        """
        raise NotImplementedError

    @abstractmethod
    def summary(self) -> str:
        """
        Get string including summary of the dataset
        """
        raise NotImplementedError

    @abstractmethod
    def get_multi(self, items: Optional[Sequence[Union[int, Hashable]]] = None, *args) -> List[Dict]:
        """
        Get multiple items, optionally just some of the keys
        :param items: specify the list of sequence to read or None for all
        """
        raise NotImplementedError



import unittest
from testbook import testbook
from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess


def run_notebook() -> None:
    notebook_path = "examples/fuse_examples/imaging/hello_world/hello_world.ipynb"

    # Execute the whole notebook and save it as an object
    with testbook(notebook_path, execute=True, timeout=600) as tb:
        # Sanity check
        test_result_acc = tb.ref("test_result_acc")
        assert test_result_acc > 0.95


@unittest.skip("until we will fix import issue")
class NotebookHelloWorldTestCase(unittest.TestCase):
    def test_notebook(self):
        run_in_subprocess(run_notebook)


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os
import copy
import logging
from typing import OrderedDict

import torch
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data.dataloader import DataLoader

from fuse.dl.models import ModelMultiHead
from fuse.dl.models.backbones.backbone_resnet import BackboneResnet
from fuse.dl.models.heads.head_global_pooling_classifier import HeadGlobalPoolingClassifier
from fuse.dl.models.backbones.backbone_inception_resnet_v2 import BackboneInceptionResnetV2

from fuse.eval.metrics.classification.metrics_thresholding_common import MetricApplyThresholds
from fuse.eval.metrics.classification.metrics_classification_common import MetricAccuracy, MetricAUCROC, MetricROCCurve

from fuse.utils.utils_debug import FuseDebug
from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.file_io.file_io import create_dir, save_dataframe, load_pickle
import fuse.utils.gpu as GPU

from fuse.eval.evaluator import EvaluatorDefault
from fuse.dl.losses.loss_default import LossDefault

from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.split import dataset_balanced_division_to_folds

import pytorch_lightning as pl
from fuse.dl.lightning.pl_module import LightningModuleDefault
from fuse.dl.lightning.pl_funcs import convert_predictions_to_dataframe

from fuseimg.datasets.isic import ISIC
from fuse_examples.imaging.classification.isic.golden_members import FULL_GOLDEN_MEMBERS


###########################################################################################################
# Fuse
###########################################################################################################
##########################################
# Debug modes
##########################################
mode = "default"  # Options: 'default', 'debug'. See details in FuseDebug
debug = FuseDebug(mode)

##########################################
# Output Paths
##########################################

NUM_GPUS = 1
ROOT = "_examples/isic/"
model_dir = os.path.join(ROOT, "model_dir")
PATHS = {
    "model_dir": model_dir,
    "inference_dir": os.path.join(model_dir, "infer_dir"),
    "eval_dir": os.path.join(model_dir, "eval_dir"),
    "data_dir": os.path.join(ROOT, "data_dir"),
    "cache_dir": os.path.join(ROOT, "cache_dir"),
    "data_split_filename": os.path.join(ROOT, "isic_split.pkl"),
}

##########################################
# Train Common Params
##########################################
TRAIN_COMMON_PARAMS = {}
# ============
# Data
# ============
TRAIN_COMMON_PARAMS["data.batch_size"] = 8
TRAIN_COMMON_PARAMS["data.train_num_workers"] = 8
TRAIN_COMMON_PARAMS["data.validation_num_workers"] = 8
TRAIN_COMMON_PARAMS["data.num_folds"] = 5
TRAIN_COMMON_PARAMS["data.train_folds"] = [0, 1, 2]
TRAIN_COMMON_PARAMS["data.validation_folds"] = [3]
TRAIN_COMMON_PARAMS["data.samples_ids"] = FULL_GOLDEN_MEMBERS  # Change to None to use all members

# ===============
# PL Trainer
# ===============
TRAIN_COMMON_PARAMS["trainer.num_epochs"] = 20
TRAIN_COMMON_PARAMS["trainer.num_devices"] = NUM_GPUS
TRAIN_COMMON_PARAMS["trainer.accelerator"] = "gpu"
TRAIN_COMMON_PARAMS["trainer.ckpt_path"] = None

# ===============
# Optimizer
# ===============
TRAIN_COMMON_PARAMS["opt.lr"] = 1e-5
TRAIN_COMMON_PARAMS["opt.weight_decay"] = 1e-3

# ===============
# Model
# ===============
TRAIN_COMMON_PARAMS["model"] = dict(dropout_rate=0.5)


def create_model(dropout_rate: float) -> torch.nn.Module:
    """
    creates the model
    """
    model = ModelMultiHead(
        conv_inputs=(("data.input.img", 3),),
        backbone={
            "Resnet18": BackboneResnet(pretrained=True, in_channels=3, name="resnet18"),
            "InceptionResnetV2": BackboneInceptionResnetV2(input_channels_num=3, logical_units_num=43),
        }["InceptionResnetV2"],
        heads=[
            HeadGlobalPoolingClassifier(
                head_name="head_0",
                dropout_rate=dropout_rate,
                conv_inputs=[("model.backbone_features", 1536)],
                num_classes=8,
                pooling="avg",
            ),
        ],
    )
    return model


#################################
# Train Template
#################################
def run_train(paths: dict, train_common_params: dict) -> None:
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Train", {"attrs": ["bold", "underline"]})

    lgr.info(f'model_dir={paths["model_dir"]}', {"color": "magenta"})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {"color": "magenta"})

    # ==============================================================================
    # Data
    # ==============================================================================
    # Train Data
    lgr.info("Train Data:", {"attrs": "bold"})

    # split to folds randomly - temp
    all_dataset = ISIC.dataset(
        paths["data_dir"],
        paths["cache_dir"],
        reset_cache=False,
        num_workers=train_common_params["data.train_num_workers"],
        samples_ids=train_common_params["data.samples_ids"],
    )

    folds = dataset_balanced_division_to_folds(
        dataset=all_dataset,
        output_split_filename=paths["data_split_filename"],
        keys_to_balance=["data.label"],
        nfolds=train_common_params["data.num_folds"],
    )

    train_sample_ids = []
    for fold in train_common_params["data.train_folds"]:
        train_sample_ids += folds[fold]
    validation_sample_ids = []
    for fold in train_common_params["data.validation_folds"]:
        validation_sample_ids += folds[fold]

    train_dataset = ISIC.dataset(paths["data_dir"], paths["cache_dir"], samples_ids=train_sample_ids, train=True)

    lgr.info("- Create sampler:")
    sampler = BatchSamplerDefault(
        dataset=train_dataset,
        balanced_class_name="data.label",
        num_balanced_classes=8,
        batch_size=train_common_params["data.batch_size"],
    )
    lgr.info("- Create sampler: Done")

    # Create dataloader
    train_dataloader = DataLoader(
        dataset=train_dataset,
        batch_sampler=sampler,
        collate_fn=CollateDefault(),
        num_workers=train_common_params["data.train_num_workers"],
    )

    lgr.info("Train Data: Done", {"attrs": "bold"})

    ## Validation data
    lgr.info("Validation Data:", {"attrs": "bold"})

    # dataset
    validation_dataset = ISIC.dataset(
        paths["data_dir"], paths["cache_dir"], samples_ids=validation_sample_ids, train=False
    )

    # dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset,
        batch_size=train_common_params["data.batch_size"],
        collate_fn=CollateDefault(),
        num_workers=train_common_params["data.validation_num_workers"],
    )
    lgr.info("Validation Data: Done", {"attrs": "bold"})

    # ==============================================================================
    # Model
    # ==============================================================================
    lgr.info("Model:", {"attrs": "bold"})

    model = create_model(**train_common_params["model"])

    lgr.info("Model: Done", {"attrs": "bold"})

    # ====================================================================================
    #  Loss
    # ====================================================================================
    losses = {
        "cls_loss": LossDefault(pred="model.logits.head_0", target="data.label", callable=F.cross_entropy, weight=1.0),
    }

    # ====================================================================================
    # Metrics
    # ====================================================================================
    class_names = ["MEL", "NV", "BCC", "AK", "BKL", "DF", "VASC", "SCC"]
    train_metrics = OrderedDict(
        [
            ("op", MetricApplyThresholds(pred="model.output.head_0")),  # will apply argmax
            ("auc", MetricAUCROC(pred="model.output.head_0", target="data.label", class_names=class_names)),
            ("accuracy", MetricAccuracy(pred="results:metrics.op.cls_pred", target="data.label")),
        ]
    )

    validation_metrics = copy.deepcopy(train_metrics)  # use the same metrics in validation as well

    best_epoch_source = dict(monitor="validation.metrics.auc.macro_avg", mode="max")

    # create optimizer
    optimizer = optim.Adam(
        model.parameters(), lr=train_common_params["opt.lr"], weight_decay=train_common_params["opt.weight_decay"]
    )

    # create learning scheduler
    lr_scheduler = {
        "ReduceLROnPlateau": optim.lr_scheduler.ReduceLROnPlateau(optimizer),
        "CosineAnnealing": optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=1),
    }["ReduceLROnPlateau"]
    lr_sch_config = dict(scheduler=lr_scheduler, monitor="validation.losses.total_loss")

    # optimizier and lr sch - see pl.LightningModule.configure_optimizers return value for all options
    optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    # =====================================================================================
    #  Train
    # =====================================================================================
    lgr.info("Train:", {"attrs": "bold"})

    # create instance of PL module - FuseMedML generic version
    pl_module = LightningModuleDefault(
        model_dir=paths["model_dir"],
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=validation_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )

    # create lightining trainer.
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train_common_params["trainer.num_epochs"],
        accelerator=train_common_params["trainer.accelerator"],
        devices=train_common_params["trainer.num_devices"],
        auto_select_gpus=True,
    )

    # train
    pl_trainer.fit(
        pl_module, train_dataloader, validation_dataloader, ckpt_path=train_common_params["trainer.ckpt_path"]
    )

    lgr.info("Train: Done", {"attrs": "bold"})


######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS["infer_filename"] = "infer_file.gz"
INFER_COMMON_PARAMS["checkpoint"] = "best_epoch.ckpt"
INFER_COMMON_PARAMS["data.num_workers"] = TRAIN_COMMON_PARAMS["data.train_num_workers"]
INFER_COMMON_PARAMS["data.infer_folds"] = [4]  # infer validation set
INFER_COMMON_PARAMS["data.batch_size"] = 4

INFER_COMMON_PARAMS["model"] = TRAIN_COMMON_PARAMS["model"]
INFER_COMMON_PARAMS["trainer.num_devices"] = TRAIN_COMMON_PARAMS["trainer.num_devices"]
INFER_COMMON_PARAMS["trainer.accelerator"] = TRAIN_COMMON_PARAMS["trainer.accelerator"]

######################################
# Inference Template
######################################


def run_infer(paths: dict, infer_common_params: dict) -> None:
    create_dir(paths["inference_dir"])
    infer_file = os.path.join(paths["inference_dir"], infer_common_params["infer_filename"])
    checkpoint_file = os.path.join(paths["model_dir"], infer_common_params["checkpoint"])

    ## Logger
    fuse_logger_start(output_path=paths["inference_dir"], console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Inference", {"attrs": ["bold", "underline"]})
    lgr.info(f"infer_filename={infer_file}", {"color": "magenta"})

    ## Data
    folds = load_pickle(paths["data_split_filename"])  # assume exists and created in train func

    infer_sample_ids = []
    for fold in infer_common_params["data.infer_folds"]:
        infer_sample_ids += folds[fold]

    # Create dataset
    infer_dataset = ISIC.dataset(paths["data_dir"], paths["cache_dir"], samples_ids=infer_sample_ids, train=False)

    # dataloader
    infer_dataloader = DataLoader(
        dataset=infer_dataset,
        collate_fn=CollateDefault(),
        batch_size=infer_common_params["data.batch_size"],
        num_workers=infer_common_params["data.num_workers"],
    )

    # load python lightning module
    model = create_model(**infer_common_params["model"])
    pl_module = LightningModuleDefault.load_from_checkpoint(
        checkpoint_file, model_dir=paths["model_dir"], model=model, map_location="cpu", strict=True
    )
    # set the prediction keys to extract (the ones used be the evaluation function).
    pl_module.set_predictions_keys(["model.output.head_0", "data.label"])  # which keys to extract and dump into file

    # create a trainer instance
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        accelerator=infer_common_params["trainer.accelerator"],
        devices=infer_common_params["trainer.num_devices"],
        auto_select_gpus=True,
    )
    predictions = pl_trainer.predict(pl_module, infer_dataloader, return_predictions=True)

    # convert list of batch outputs into a dataframe
    infer_df = convert_predictions_to_dataframe(predictions)
    save_dataframe(infer_df, infer_file)


######################################
# Eval Common Params
######################################
EVAL_COMMON_PARAMS = {}
EVAL_COMMON_PARAMS["infer_filename"] = INFER_COMMON_PARAMS["infer_filename"]


######################################
# Eval Template
######################################
def run_eval(paths: dict, eval_common_params: dict) -> None:
    infer_file = os.path.join(paths["inference_dir"], eval_common_params["infer_filename"])

    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Eval", {"attrs": ["bold", "underline"]})

    # metrics
    metrics = OrderedDict(
        [
            ("op", MetricApplyThresholds(pred="model.output.head_0")),  # will apply argmax
            ("auc", MetricAUCROC(pred="model.output.head_0", target="data.label")),
            ("accuracy", MetricAccuracy(pred="results:metrics.op.cls_pred", target="data.label")),
            (
                "roc",
                MetricROCCurve(
                    pred="model.output.head_0",
                    target="data.label",
                    output_filename=os.path.join(paths["inference_dir"], "roc_curve.png"),
                ),
            ),
        ]
    )

    # create evaluator
    evaluator = EvaluatorDefault()

    # run
    results = evaluator.eval(ids=None, data=infer_file, metrics=metrics, output_dir=paths["eval_dir"])

    return results


######################################
# Run
######################################
if __name__ == "__main__":
    ## allocate gpus
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    GPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    ISIC.download(data_path=PATHS["data_dir"])

    RUNNING_MODES = ["train", "infer", "eval"]  # Options: 'train', 'infer', 'eval'

    # train
    if "train" in RUNNING_MODES:
        run_train(paths=PATHS, train_common_params=TRAIN_COMMON_PARAMS)

    # infer
    if "infer" in RUNNING_MODES:
        run_infer(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # eval
    if "eval" in RUNNING_MODES:
        run_eval(paths=PATHS, eval_common_params=EVAL_COMMON_PARAMS)



import functools
from typing import Any, List, Optional
from fuse.utils.utils_debug import FuseDebug
import torch
from tqdm import tqdm
import multiprocessing as mp
from termcolor import cprint
import os
import traceback

"""
global dictionary that stores arguments to a new created process
Direct access is not allowed - use get_from_global_storage to access it from a worker function.
call the following "private" functions only if you know what you're doing: _store_in_global_storage, _remove_from_global_storage
Typically, you'd only need to call get_from_global_storage from your worker_func, and call run_multiprocessed and provide copy_global_storage to it,
    with a dict of values that you want accesible from the worker_func.
"""
_multiprocess_global_storage = {}


def __orig__run_multiprocessed(
    worker_func,
    args_list,
    workers=0,
    verbose=0,
    copy_to_global_storage: Optional[dict] = None,
    keep_results_order: bool = True,
) -> List[Any]:
    """
    Args:
        worker_func: a worker function, must accept only a single positional argument and no optional args.
            For example:
            def some_worker(args):
                speed, height, banana = args
                ...
                return ans
        args_list: a list in which each element is the input to func
        workers: number of processes to use. Use 0 for no spawning of processes (helpful when debugging)
        copy_to_global_storage: Optional - to optimize the running time - the provided dict will be stored in a way that is accesible to worker_func.
         calling get_from_global_storage(...) will allow access to it from within any worker_func
        This allows to create a significant speedup in certain cases, and the main idea is that it allows to drastically reduce the amount of data
         that gets (automatically) pickled by python's multiprocessing library.
        Instead of copying it for each worker_func invocation, it will be copied once, upon worker process initialization.
    Returns:
        List of results from calling func, in the same order as args_list
    """
    if "DEBUG_SINGLE_PROCESS" in os.environ and os.environ["DEBUG_SINGLE_PROCESS"] in ["T", "t", "True", "true", 1]:
        workers = None
        cprint(
            "Due to the env variable DEBUG_SINGLE_PROCESS being set, run_multiprocessed is not using multiprocessing",
            "red",
        )
    assert callable(worker_func)

    if verbose < 1:
        tqdm_func = lambda x: x
    else:
        tqdm_func = tqdm

    if copy_to_global_storage is None:
        copy_to_global_storage = {}

    all_res = []
    if workers is None or workers <= 1:
        _store_in_global_storage(copy_to_global_storage)
        for i in tqdm_func(range(len(args_list))):
            curr_ans = worker_func(args_list[i])
            all_res.append(curr_ans)
        _remove_from_global_storage(list(copy_to_global_storage.keys()))
    else:
        assert isinstance(workers, int)
        assert workers >= 0

        with mp.Pool(
            processes=workers,
            initializer=_store_in_global_storage,
            initargs=(copy_to_global_storage,),
            maxtasksperchild=400,
        ) as pool:
            if verbose > 0:
                cprint(f"multiprocess pool created with {workers} workers.", "cyan")
            map_func = pool.imap if keep_results_order else pool.imap_unordered
            for curr_ans in tqdm_func(
                map_func(worker_func, args_list), total=len(args_list), smoothing=0.1, disable=verbose < 1
            ):
                all_res.append(curr_ans)

    return all_res


def run_multiprocessed(
    worker_func,
    args_list,
    workers=0,
    verbose=0,
    copy_to_global_storage: Optional[dict] = None,
    keep_results_order: bool = True,
    as_iterator=False,
    mp_context: Optional[str] = None,
) -> List[Any]:
    """
    Args:
        worker_func: a worker function, must accept only a single positional argument and no optional args.
            For example:
            def some_worker(args):
                speed, height, banana = args
                ...
                return ans
        args_list: a list in which each element is the input to func
        workers: number of processes to use. Use 0 for no spawning of processes (helpful when debugging)
        copy_to_global_storage: Optional - to optimize the running time - the provided dict will be stored in a way that is accessible to worker_func.
         calling get_from_global_storage(...) will allow access to it from within any worker_func
        This allows to create a significant speedup in certain cases, and the main idea is that it allows to drastically reduce the amount of data
         that gets (automatically) pickled by python's multiprocessing library.
        Instead of copying it for each worker_func invocation, it will be copied once, upon worker process initialization.
        keep_results_order: determined if imap or imap_unordered is used. if strict_answers_order is set to False, then results will be ordered by their readiness.
         if strict_answers_order is set to True, the answers will be provided at the same order as defined in the args_list
        as_iterator: if True, a lightweight iterator is returned. This is useful in the cases that the entire returned answer doesn't fit in memory.
         or in the case that you want to parallelize some calculation with the generation.
         if False, the answers will be accumulated to a list and returned.
    :param mp_context: "fork", "spawn", "thread" or None for multiprocessing default
    Returns:
        if as_iterator is set to True, returns an iterator.
        Otherwise, returns a list of results from calling func
    """

    iter = _run_multiprocessed_as_iterator_impl(
        worker_func=worker_func,
        args_list=args_list,
        workers=workers,
        verbose=verbose,
        copy_to_global_storage=copy_to_global_storage,
        keep_results_order=keep_results_order,
        mp_context=mp_context,
    )

    if as_iterator:
        return iter

    ans = [x for x in iter]
    return ans


def _run_multiprocessed_as_iterator_impl(
    worker_func,
    args_list,
    workers=0,
    verbose=0,
    copy_to_global_storage: Optional[dict] = None,
    keep_results_order: bool = True,
    mp_context: Optional[str] = None,
) -> List[Any]:
    """
    an iterator version of run_multiprocessed - useful when the accumulated answer is too large to fit in memory

    Args:
        worker_func: a worker function, must accept only a single positional argument and no optional args.
            For example:
            def some_worker(args):
                speed: height, banana = args
                ...
                return ans
        args_list: a list in which each element is the input to func
        workers: number of processes to use. Use 0 for no spawning of processes (helpful when debugging)
        copy_to_global_storage: Optional - to optimize the running time - the provided dict will be stored in a way that is accessible to worker_func.
            calling get_from_global_storage(...) will allow access to it from within any worker_func
        This allows to create a significant speedup in certain cases, and the main idea is that it allows to drastically reduce the amount of data
            that gets (automatically) pickled by python's multiprocessing library.
        Instead of copying it for each worker_func invocation, it will be copied once, upon worker process initialization.
        keep_results_order: determined if imap or imap_unordered is used. if strict_answers_order is set to False, then results will be ordered by their readiness.
            if strict_answers_order is set to True, the answers will be provided at the same order as defined in the args_list
    :param mp_context: "fork", "spawn", "thread" or None for multiprocessing default
    """
    if "DEBUG_SINGLE_PROCESS" in os.environ and os.environ["DEBUG_SINGLE_PROCESS"] in ["T", "t", "True", "true", 1]:
        workers = None
        cprint(
            "Due to the env variable DEBUG_SINGLE_PROCESS being set, run_multiprocessed is not using multiprocessing",
            "red",
        )

    if FuseDebug().get_setting("multiprocessing") == "main_process":
        workers = None
        cprint("Due to the FuseDebug mode, run_multiprocessed is not using multiprocessing", "red")

    assert callable(worker_func)

    if verbose < 1:
        tqdm_func = lambda x: x
    else:
        tqdm_func = tqdm

    if copy_to_global_storage is None:
        copy_to_global_storage = {}

    if workers is None or workers <= 1:
        _store_in_global_storage(copy_to_global_storage)
        try:
            for i in tqdm_func(range(len(args_list))):
                curr_ans = worker_func(args_list[i])
                yield curr_ans
        except:
            raise
        finally:
            _remove_from_global_storage(list(copy_to_global_storage.keys()))
    else:
        assert isinstance(workers, int)
        assert workers >= 0

        if mp_context == "thread":
            from multiprocessing.pool import ThreadPool

            pool = ThreadPool
        elif mp_context is None:  # os default
            pool = mp.Pool
        else:
            pool = mp.get_context(mp_context).Pool

        worker_func = functools.partial(worker_func_wrapper, worker_func=worker_func)
        with pool(processes=workers, initializer=_store_in_global_storage, initargs=(copy_to_global_storage,)) as pool:
            if verbose > 0:
                cprint(f"multiprocess pool created with {workers} workers.", "cyan")
            map_func = pool.imap if keep_results_order else pool.imap_unordered
            for curr_ans in tqdm_func(
                map_func(worker_func, args_list), total=len(args_list), smoothing=0.1, disable=verbose < 1
            ):
                yield curr_ans


def worker_func_wrapper(*args, worker_func, **kwargs):
    torch.set_num_threads(1)
    return worker_func(*args, **kwargs)


def _store_in_global_storage(store_me: dict) -> None:
    """
    Copy elements to new pool processes to optimize the running time
    The arguments will be added to a global dictionary multiprocess_copied_args
    :param kwargs: list of tuples - each tuple is a key-value pair and will be added to the global dictionary
    :return: None
    """
    if store_me is None:
        return

    global _multiprocess_global_storage

    # making sure there are no name conflicts
    for key, _ in store_me.items():
        assert (
            key not in _multiprocess_global_storage
        ), f"run_multiprocessed - two multiprocessing pools with num_workers=0 are running simultaneously and using the same argument name {key}"

    _multiprocess_global_storage.update(store_me)


def _remove_from_global_storage(remove_me: List) -> None:
    """
    remove copied args for multiprocess
    :param kwargs: list of tuples - each tuple is a key-value pair and will be added to the global dictionary
    :return: None
    """
    if remove_me is None:
        return

    global _multiprocess_global_storage
    for key in remove_me:
        del _multiprocess_global_storage[key]


def get_from_global_storage(key: str) -> Any:
    """
    Get args copied by run_multiprocessed
    """
    global _multiprocess_global_storage
    return _multiprocess_global_storage[key]


ctx = mp.get_context("spawn")


class Process(ctx.Process):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pconn, self._cconn = mp.Pipe()
        self._start_method = None  # don't force spawn from now on

    def run(self):
        try:
            results = self._target(*self._args, **self._kwargs)
            self._cconn.send((results, None))
        except Exception as e:
            tb = traceback.format_exc()
            self._cconn.send((None, (e, tb)))
            raise e  # You can still rise this exception if you need to

    @property
    def results_and_error(self):
        if self._pconn.poll():
            return self._pconn.recv()
        return (None, None)


def run_in_subprocess(f: callable, *args, timeout: int = 600, **kwargs):
    """A decorator that makes function run in a subprocess.
    This can be useful when you want allocate GPU and memory and to release it when you're done.
    :param f: the function to run in a subprocess
    :param timeout: the maximum time to wait for the process to complete
    """
    if not os.environ.get("FORCE_RUN_IN_SUBPROCESS", True):
        return f(*args, **kwargs)

    # create process
    p = Process(target=f, args=args, kwargs=kwargs)  # using a subclass of Process
    p.start()
    try:
        p.join(timeout=timeout)
    except:
        p.terminate()
        raise

    results, error = p.results_and_error
    if error is not None:
        exception, traceback = error
        print(f"process func {f} had an exception: {exception}")
        print(traceback)
        raise RuntimeError(f"process func {f} had an exception: {exception}")

    assert p.exitcode == 0, f"process func {f} failed with exit code {p.exitcode}"
    return results



import os
from typing import Callable, Optional, Tuple
from fuse.data.ops.ops_cast import Cast
from fuse.data.utils.sample import get_sample_id
from fuse.utils import NDict
from fuse.utils.file_io.file_io import create_dir
from fuse.data.ops.ops_debug import OpDebugBase
import numpy
import torch
import matplotlib.pyplot as plt


class OpVis2DImage(OpDebugBase):
    """
    Visualize a 2D image (either display it or save it as an image).
    Supports both rgb and gray scale and supports few image formats "channels_first"/"no_channels"/"channels_last"
    It's recommended, but not a must, to run it in a single process.
    Add at the top your script to force single process:
    ```
    from fuse.utils.utils_debug import FuseDebug
    FuseDebug("debug")
    ```
    Example:
    # Display RGB image [3, H, W] already in range [0-256]
    (OpVis2DImage(num_samples=1), dict(key="data.input.img", dtype="int"))

    """

    def __init__(
        self,
        show: bool = True,
        path: str = ".",
        image_format: str = "channels_first",
        image_process_func: Optional[Callable] = None,
        **kwargs,
    ):
        """
        :param show: if set to true will display it, otherwise will save it in `path`
        :param path: location to save the images. Used when `show` set to False
        :param image_format: the format of the image: "channels_first"/"no_channels"/"channels_last"
        :param image_processing_func: callable that get the image (numpy array) and can process it. Typically use to move the range to values matplotlib can display.
        :param kwargs: see super class arguments
        """
        super().__init__(**kwargs)
        self._path = path
        self._image_format = image_format
        self._image_process_func = image_process_func
        self._show = show

    def call_debug(
        self,
        sample_dict: NDict,
        key: str,
        clip: Optional[Tuple] = None,
        dtype: Optional[str] = None,
        figure_kwargs: dict = {},
        **imshow_kwargs,
    ) -> None:
        """
        :param key: sample_dict key to a 2D image. Either tensor or numpy array.
        :param clip: tuple of arguments to numpy.clip if the operation required
        :param dtype: in case the range is [0-256], matplotlib expects int. Set to "int" if this is the case.
        :param figure_kwargs: parameters to pass to plt.figure
        :params imshow_kwargs: extra parameters to pass to plt.imshow
        """

        img = sample_dict[key]
        if isinstance(img, torch.Tensor):
            img = img.detach().cpu().numpy()

        assert isinstance(img, numpy.ndarray)

        if clip is not None:
            img = img.clip(*clip)

        if dtype is not None:
            img = Cast.to_numpy(img, dtype=dtype)

        if self._image_format == "channels_first":
            img = numpy.moveaxis(img, 0, -1)
        elif self._image_format == "no_channels":
            img = numpy.expand_dims(img, axis=-1)

        if self._image_process_func is not None:
            img = self._image_process_func(img)

        if "cmap" not in imshow_kwargs:
            if img.shape[-1] == 1:  # assume and force gray scale
                imshow_kwargs["cmap"] = "gray"

        plt.figure(**figure_kwargs)
        plt.imshow(img, **imshow_kwargs)
        if self._show:
            plt.show()
        else:
            if self._name is None:
                filename = os.path.join(self._path, get_sample_id(sample_dict).replace(".", "__")) + ".png"
            else:
                filename = os.path.join(self._path, self._name, get_sample_id(sample_dict).replace(".", "__")) + ".png"
            create_dir(os.path.dirname(filename))
            plt.savefig(filename)  # save fig


class OpVisImageHist(OpDebugBase):
    """
    Visualize single image histogram (either display it or save it as an image).
    ```
    from fuse.utils.utils_debug import FuseDebug
    FuseDebug("debug")
    ```
    Example:
    ```
    (OpVisImageHist(first_sample_only=True), dict(key="data.input.img"))
    ```
    """

    def __init__(self, show: bool = True, path: str = ".", **kwargs):
        """
        :param show: if set to true will display it, otherwise will save it in `path`
        :param path: location to save the images. Used when `show` set to False
        :param kwargs: see super class arguments
        """
        super().__init__(**kwargs)
        self._path = path
        self._show = show

    def call_debug(
        self, sample_dict: NDict, key: str, bins: int = 10, figure_kwargs: dict = {}, **hist_kwargs
    ) -> NDict:
        """
        :param key: sample_dict key to a 2D image. Either tensor or numpy array.
        :param figure_kwargs: parameters to pass to plt.figure
        :params hist_kwargs: extra parameters to pass to plt.hist
        """

        img = sample_dict[key]
        if isinstance(img, torch.Tensor):
            img = img.detach().cpu().numpy()

        assert isinstance(img, numpy.ndarray)

        if self._image_format == "channels_first":
            img = numpy.moveaxis(img, 0, -1)
        if self._image_process_func is not None:
            img = self._image_process_func(img)

        plt.figure(**figure_kwargs)
        plt.hist(img.flatten(), bins=bins, **hist_kwargs)
        if self._show:
            plt.show()
        else:
            plt.savefig(
                os.path.join(self._path, f"hist_{get_sample_id(sample_dict).replace('.', '__')}.png")
            )  # save fig



import unittest

from typing import Optional, OrderedDict, Union, List

from fuse.utils.ndict import NDict

from fuse.data.ops.op_base import OpBase, OpReversibleBase, op_call
from fuse.data.key_types_for_testing import DataTypeForTesting

from fuse.data.ops.ops_common import OpApplyPatterns, OpFunc, OpLambda, OpRepeat
from fuse.data.ops.ops_common_for_testing import OpApplyTypesImaging


class OpIncrForTest(OpReversibleBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(
        self, sample_dict: NDict, op_id: Optional[str], incr_value: int, key_in: str, key_out: str
    ) -> Union[None, dict, List[dict]]:
        # save for reverse
        sample_dict[op_id] = {"key_out": key_out, "incr_value": incr_value}
        # apply
        value = sample_dict[key_in]
        sample_dict[key_out] = value + incr_value

        return sample_dict

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        # not really reverse, but help the test
        orig_args = sample_dict[op_id]

        if orig_args["key_out"] != key_to_follow:
            return sample_dict

        value = sample_dict[key_to_reverse]
        sample_dict[key_to_reverse] = value - orig_args["incr_value"]

        return sample_dict


class TestOpsCommon(unittest.TestCase):
    def test_op_repeat(self):
        """
        Test OpRepeat __call__() and reverse()
        """
        op_base = OpIncrForTest()
        kwargs_per_step_to_add = [
            dict(key_in="data.val.a", key_out="data.val.b"),
            dict(key_in="data.val.b", key_out="data.val.c"),
            dict(key_in="data.val.b", key_out="data.val.d"),
            dict(key_in="data.val.d", key_out="data.val.d"),
        ]
        op_repeat = OpRepeat(op_base, kwargs_per_step_to_add)
        sample_dict = NDict({})
        sample_dict["data.val.a"] = 5
        sample_dict = op_repeat(sample_dict, "_.test_repeat", incr_value=3)
        self.assertEqual(sample_dict["data.val.a"], 5)
        self.assertEqual(sample_dict["data.val.b"], 8)
        self.assertEqual(sample_dict["data.val.c"], 11)
        self.assertEqual(sample_dict["data.val.d"], 14)

        op_repeat.reverse(sample_dict, key_to_follow="data.val.d", key_to_reverse="data.val.d", op_id="_.test_repeat")
        self.assertEqual(sample_dict["data.val.a"], 5)
        self.assertEqual(sample_dict["data.val.b"], 8)
        self.assertEqual(sample_dict["data.val.c"], 11)
        self.assertEqual(sample_dict["data.val.d"], 8)

        sample_dict["data.val.e"] = 48
        op_repeat.reverse(sample_dict, key_to_follow="data.val.d", key_to_reverse="data.val.e", op_id="_.test_repeat")
        self.assertEqual(sample_dict["data.val.a"], 5)
        self.assertEqual(sample_dict["data.val.b"], 8)
        self.assertEqual(sample_dict["data.val.c"], 11)
        self.assertEqual(sample_dict["data.val.d"], 8)
        self.assertEqual(sample_dict["data.val.e"], 42)

    def test_op_lambda(self):
        """
        Test OpLambda __call__() and reverse()
        """
        op_base = OpLambda(func=lambda x: x + 3)
        kwargs_per_step_to_add = [dict(), dict(), dict()]
        op_repeat = OpRepeat(op_base, kwargs_per_step_to_add)
        sample_dict = NDict({})
        sample_dict["data.val.a"] = 5
        sample_dict = op_repeat(sample_dict, "_.test_repeat", key="data.val.a")
        self.assertEqual(sample_dict["data.val.a"], 14)

        op_base = OpLambda(func=lambda x: x + 3, func_reverse=lambda x: x - 3)
        op_repeat = OpRepeat(op_base, kwargs_per_step_to_add)
        sample_dict = NDict({})
        sample_dict["data.val.a"] = 5
        sample_dict = op_repeat(sample_dict, "_.test_repeat", key="data.val.a")
        self.assertEqual(sample_dict["data.val.a"], 14)

        op_repeat.reverse(sample_dict, key_to_follow="data.val.a", key_to_reverse="data.val.a", op_id="_.test_repeat")
        self.assertEqual(sample_dict["data.val.a"], 5)

        sample_dict["data.val.b"] = 51
        op_repeat.reverse(sample_dict, key_to_follow="data.val.a", key_to_reverse="data.val.b", op_id="_.test_repeat")
        self.assertEqual(sample_dict["data.val.a"], 5)
        self.assertEqual(sample_dict["data.val.b"], 42)

    def test_op_lambda_with_kwargs(self):
        """
        Test OpLambda __call__() with kwargs
        """
        op_base = OpLambda(func=lambda x, y: x + y)
        kwargs_per_step_to_add = [dict(), dict(), dict()]
        op_repeat = OpRepeat(op_base, kwargs_per_step_to_add)
        sample_dict = NDict()
        sample_dict["data.val.a"] = 5
        sample_dict = op_repeat(sample_dict, "_.test_repeat", key="data.val.a", y=5)
        self.assertEqual(sample_dict["data.val.a"], 20)

    def test_op_func(self):
        """
        Test OpFunc __call__()
        """

        def func_single_output(a, b, c):
            return a + b + c

        def func_multi_output(a, b, c):
            return a + b, a + c

        single_output_op = OpFunc(func=func_single_output)
        sample_dict = NDict({})
        sample_dict["data.first"] = 5
        sample_dict["data.second"] = 9
        sample_dict = op_call(
            single_output_op,
            sample_dict,
            "_.test_func",
            c=2,
            inputs={"data.first": "a", "data.second": "b"},
            outputs="data.out",
        )
        self.assertEqual(sample_dict["data.out"], 16)

        multi_output_op = OpFunc(func=func_multi_output)
        sample_dict = NDict({})
        sample_dict["data.first"] = 5
        sample_dict["data.second"] = 9
        sample_dict = op_call(
            multi_output_op,
            sample_dict,
            "_.test_func",
            c=2,
            inputs={"data.first": "a", "data.second": "b"},
            outputs=["data.out", "data.more"],
        )
        self.assertEqual(sample_dict["data.out"], 14)
        self.assertEqual(sample_dict["data.more"], 7)

    def test_op_apply_patterns(self):
        """
        Test OpRApplyPatterns __call__() and reverse()
        """

        op_add_1 = OpLambda(func=lambda x: x + 1, func_reverse=lambda x: x - 1)
        op_mul_2 = OpLambda(func=lambda x: x * 2, func_reverse=lambda x: x // 2)
        op_mul_4 = OpLambda(func=lambda x: x * 4, func_reverse=lambda x: x // 4)

        sample_dict = NDict({})
        sample_dict["data.val.img_for_testing"] = 3
        sample_dict["data.test.img_for_testing"] = 3
        sample_dict["data.test.seg_for_testing"] = 3
        sample_dict["data.test.bbox_for_testing"] = 3
        sample_dict["data.test.meta"] = 3

        patterns_dict = OrderedDict(
            [
                (r"^data.val.img_for_testing$", (op_add_1, dict())),
                (r"^.*img_for_testing$|^.*seg_for_testing$", (op_mul_2, dict())),
                (r"^data.[^.]*.bbox_for_testing", (op_mul_4, dict())),
            ]
        )
        op_apply_pat = OpApplyPatterns(patterns_dict)

        sample_dict = op_apply_pat(sample_dict, "_.test_apply_pat")
        self.assertEqual(sample_dict["data.val.img_for_testing"], 4)
        self.assertEqual(sample_dict["data.test.img_for_testing"], 6)
        self.assertEqual(sample_dict["data.test.seg_for_testing"], 6)
        self.assertEqual(sample_dict["data.test.bbox_for_testing"], 12)
        self.assertEqual(sample_dict["data.test.meta"], 3)

        sample_dict["model.seg_for_testing"] = 3
        op_apply_pat.reverse(
            sample_dict,
            key_to_follow="data.val.img_for_testing",
            key_to_reverse="model.seg_for_testing",
            op_id="_.test_apply_pat",
        )
        self.assertEqual(sample_dict["data.val.img_for_testing"], 4)
        self.assertEqual(sample_dict["model.seg_for_testing"], 2)

    def test_op_apply_types(self):
        """
        Test OpApplyTypes __call__() and reverse()
        """

        op_add_1 = OpLambda(func=lambda x: x + 1, func_reverse=lambda x: x - 1)
        op_mul_2 = OpLambda(func=lambda x: x * 2, func_reverse=lambda x: x // 2)
        op_mul_4 = OpLambda(func=lambda x: x * 4, func_reverse=lambda x: x // 4)

        sample_dict = NDict({})
        sample_dict["data.val.img_for_testing"] = 3
        sample_dict["data.test.img_for_testing"] = 3
        sample_dict["data.test.seg_for_testing"] = 3
        sample_dict["data.test.bbox_for_testing"] = 3
        sample_dict["data.test.meta"] = 3

        types_dict = {
            DataTypeForTesting.IMAGE_FOR_TESTING: (op_add_1, dict()),
            DataTypeForTesting.SEG_FOR_TESTING: (op_mul_2, dict()),
            DataTypeForTesting.BBOX_FOR_TESTING: (op_mul_4, dict()),
        }

        op_apply_type = OpApplyTypesImaging(types_dict)

        sample_dict = op_apply_type(sample_dict, "_.test_apply_type")
        self.assertEqual(sample_dict["data.val.img_for_testing"], 4)
        self.assertEqual(sample_dict["data.test.img_for_testing"], 4)
        self.assertEqual(sample_dict["data.test.seg_for_testing"], 6)
        self.assertEqual(sample_dict["data.test.bbox_for_testing"], 12)
        self.assertEqual(sample_dict["data.test.meta"], 3)

        sample_dict["model.a_seg_for_testing"] = 3
        op_apply_type.reverse(
            sample_dict,
            key_to_follow="data.val.img_for_testing",
            key_to_reverse="model.a_seg_for_testing",
            op_id="_.test_apply_type",
        )
        self.assertEqual(sample_dict["data.val.img_for_testing"], 4)
        self.assertEqual(sample_dict["model.a_seg_for_testing"], 2)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from fuse.data.key_types import DataTypeBasic
from fuse.data.patterns import Patterns


class TypedElement:
    """
    encapsulates a single item view with all its overlayed data
    """

    def __init__(self, image=None, seg=None, contours=None, bboxes=None, labels=None, metadata=None) -> None:
        assert isinstance(image, (np.ndarray, type(None)))
        assert isinstance(seg, (np.ndarray, type(None)))
        # assert isinstance(contours, (np.ndarray, type(None)))
        # assert isinstance(bboxes, (np.ndarray, type(None)))
        # assert isinstance(labels, (np.ndarray, type(None)))

        self.image = image
        self.seg = seg
        self.contours = contours
        self.bboxes = bboxes
        self.labels = labels
        self.metadata = metadata


def typedElementFromSample(sample_dict, key_pattern, td):
    patterns = Patterns({key_pattern: True}, False)
    all_keys = [k for k in sample_dict.get_all_keys() if patterns.get_value(k)]

    content = {
        td.get_type(sample_dict, k).value: sample_dict[k]
        for k in all_keys
        if td.get_type(sample_dict, k) != DataTypeBasic.UNKNOWN
    }
    keymap = {td.get_type(sample_dict, k): k for k in all_keys if td.get_type(sample_dict, k) != DataTypeBasic.UNKNOWN}
    elem = TypedElement(**content)
    return elem, keymap


def typedElementToSample(sample_dict, typed_element, keymap):
    for k, v in keymap.items():
        sample_dict[v] = typed_element.__getattribute__(k.value)
    return sample_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Hashable, List, Optional, Dict, Union
from fuse.utils.file_io.file_io import read_dataframe
import pandas as pd

from fuse.data import OpBase
from fuse.utils.ndict import NDict


class OpReadDataframe(OpBase):
    """
    Op reading data from pickle file / dataframe object.
    Each row will be added as a value to sample dict.
    """

    def __init__(
        self,
        data: Optional[pd.DataFrame] = None,
        data_filename: Optional[str] = None,
        columns_to_extract: Optional[List[str]] = None,
        rename_columns: Optional[Dict[str, str]] = None,
        key_name: str = "data.sample_id",
        key_column: str = "sample_id",
    ):
        """
        :param data:  input DataFrame
        :param data_filename: path to a pickled DataFrame (possible zipped)
        :param columns_to_extract: list of columns to extract from dataframe. When None (default) all columns are extracted
        :param rename_columns: rename columns from dataframe, when None (default) column names are kept
        :param key_name: name of value in sample_dict which will be used as the key/index
        :param key_column: name of the column which use as key/index
        """
        super().__init__()

        # store input
        self._data_filename = data_filename
        self._columns_to_extract = columns_to_extract
        self._rename_columns = rename_columns
        self._key_name = key_name
        self._key_column = key_column
        df = data

        # verify input
        if data is None and data_filename is None:
            msg = "Error: need to provide either in-memory DataFrame or a path to file."
            raise Exception(msg)
        elif data is not None and data_filename is not None:
            msg = "Error: need to provide either 'data' or 'data_filename' args, bot not both."
            raise Exception(msg)

        # read dataframe
        if self._data_filename is not None:
            df = read_dataframe(self._data_filename)

        # extract only specified columns (in case not specified, extract all)
        if self._columns_to_extract is not None:
            df = df[self._columns_to_extract]

        # rename columns
        if self._rename_columns is not None:
            df.rename(self._rename_columns, axis=1, inplace=True)

        # convert to dictionary: {index -> {column -> value}}
        if self._key_column is not None:
            df = df.set_index(self._key_column)
        self._data = df.to_dict(orient="index")

    def __call__(self, sample_dict: NDict, prefix: Optional[str] = None, **kwargs) -> Union[None, dict, List[dict]]:
        """
        See base class

        :param prefix: specify a prefix for the sample dict keys.
                       For example, with prefix 'data.features' and a df with the columns ['height', 'weight', 'sex'],
                       the matching keys will be: 'data.features.height', 'data.features.weight', 'data.features.sex'.
        """
        key = sample_dict[self._key_name]

        # locate the required item
        sample_data = self._data[key].copy()

        # add values tp sample_dict
        for name, value in sample_data.items():
            if prefix is None:
                sample_dict[name] = value
            else:
                sample_dict[f"{prefix}.{name}"] = value

        return sample_dict

    def get_all_keys(self) -> List[Hashable]:
        """
        :return: list of  dataframe index values
        """
        return list(self.data.keys())



import zipfile
from tqdm import tqdm
import os


def extract_zip_file(filename: str, dest_dir: str, show_progress_bar=True):
    os.makedirs(dest_dir, exist_ok=True)
    print(f"extracting {filename} ...")
    with zipfile.ZipFile(filename) as zf:
        if show_progress_bar:
            for member in tqdm(zf.infolist(), desc="Extracting "):
                zf.extract_zip_file(member, dest_dir)
        else:
            zf.extractall(dest_dir)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021
"""

import pytorch_lightning as pl
from typing import Optional
from fuse.dl.lightning.pl_funcs import *  # noqa


class LightningModuleDefault(pl.LightningModule):
    """
    Generic implementation of LightningModule using FuseMedML style focusing primarily on supervised training.
    FuseMedML conventions make it possible to have such a generic implementation.
    """

    def __init__(
        self,
        model_dir: str,
        model: Optional[torch.nn.Module] = None,
        losses: Optional[Dict[str, LossBase]] = None,
        train_metrics: Optional[OrderedDict[str, MetricBase]] = None,
        validation_metrics: Optional[OrderedDict[str, MetricBase]] = None,
        test_metrics: Optional[OrderedDict[str, MetricBase]] = None,
        optimizers_and_lr_schs: Any = None,
        callbacks: Optional[Sequence[pl.Callback]] = None,
        best_epoch_source: Optional[Union[Dict, List[Dict]]] = None,
        save_hyperparameters: Optional[List[str]] = None,
        **kwargs
    ):
        """
        :param model_dir: location for checkpoints and logs
        :param model: Pytorch model to use
        :param optimizers_and_lr_schs: see pl.LightningModule.configure_optimizers for details and relevant options
        :param losses: dict of FuseMedML style losses
        :param train_metrics: dict of FuseMedML style metrics - used for trainset
        :param validation_metrics: dict of FuseMedML style metrics - used for validation set (must be different instances of metrics (from train_metrics!)
        :param test_metrics: dict of FuseMedML style metrics - used for test set (must be different instances of metrics (from train_metrics and validation_metrics!)
        :param optimizers_and_lr_schs: see pl.LightningModule.configure_optimizers return value for all options
        :param callbacks: see pl.LightningModule.configure_callbacks return value for details
        :param best_epoch_source: Create list of pl.callbacks that saves checkpoints using (pl.callbacks.ModelCheckpoint) and print per epoch summary (fuse.dl.lightning.pl_epoch_summary.ModelEpochSummary).
                                  Either a dict with arguments to pass to ModelCheckpoint or list dicts for multiple ModelCheckpoint callbacks (to monitor and save checkpoints for more then one metric).
        :param save_hyperparameters: specify which hyperparameters you would like to save. Default None.  See pl.LightningModule.save_hyperparameters() for more details.
        """
        super().__init__(**kwargs)
        if save_hyperparameters is not None:
            self.save_hyperparameters(*save_hyperparameters)

        # store arguments
        self._model_dir = model_dir
        self._model = model
        self._losses = losses if losses is not None else {}
        self._train_metrics = train_metrics if train_metrics is not None else {}
        self._validation_metrics = validation_metrics if validation_metrics is not None else {}
        self._test_metrics = test_metrics if test_metrics is not None else {}

        self._optimizers_and_lr_schs = optimizers_and_lr_schs
        self._callbacks = callbacks if callbacks is not None else []
        if best_epoch_source is not None:
            self._callbacks += model_checkpoint_callbacks(model_dir, best_epoch_source)

        # init state
        self._prediction_keys = None

    ## forward
    def forward(self, batch_dict: NDict) -> NDict:
        return self._model(batch_dict)

    ## Step
    def training_step(self, batch_dict: NDict, batch_idx: int) -> dict:
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # given the batch_dict and FuseMedML style losses - compute the losses, return the total loss and save losses values in batch_dict["losses"]
        total_loss = step_losses(self._losses, batch_dict)
        # given the batch_dict and FuseMedML style losses - collect the required values to compute the metrics on epoch_end
        step_metrics(self._train_metrics, batch_dict)

        # return the total_loss, the losses and drop everything else
        return {"loss": total_loss, "losses": batch_dict["losses"]}

    def validation_step(self, batch_dict: NDict, batch_idx: int) -> dict:
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # given the batch_dict and FuseMedML style losses - compute the losses, return the total loss (ignored) and save losses values in batch_dict["losses"]
        _ = step_losses(self._losses, batch_dict)
        # given the batch_dict and FuseMedML style losses - collect the required values to compute the metrics on epoch_end
        step_metrics(self._validation_metrics, batch_dict)

        # return just the losses and drop everything else
        return {"losses": batch_dict["losses"]}

    def test_step(self, batch_dict: NDict, batch_idx: int) -> dict:
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # given the batch_dict and FuseMedML style losses - compute the losses, return the total loss (ignored) and save losses values in batch_dict["losses"]
        _ = step_losses(self._losses, batch_dict)
        # given the batch_dict and FuseMedML style losses - collect the required values to compute the metrics on epoch_end
        step_metrics(self._test_metrics, batch_dict)

        # return just the losses and drop everything else
        return {"losses": batch_dict["losses"]}

    def predict_step(self, batch_dict: NDict, batch_idx: int) -> dict:
        if self._prediction_keys is None:
            raise Exception(
                "Error: predict_step expectes list of prediction keys to extract from batch_dict. Please specify it using set_predictions_keys() method "
            )
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # extract the requried keys - defined in self.set_predictions_keys()
        return step_extract_predictions(self._prediction_keys, batch_dict)

    ## Epoch end
    def training_epoch_end(self, step_outputs) -> None:
        # for the logs to be at each epoch, not each step
        self.log("step", self.current_epoch, on_epoch=True)
        # calc average epoch loss and log it
        epoch_end_compute_and_log_losses(self, "train", [e["losses"] for e in step_outputs])
        # evaluate  and log it
        epoch_end_compute_and_log_metrics(self, "train", self._train_metrics)

    def validation_epoch_end(self, step_outputs) -> None:
        # for the logs to be at each epoch, not each step
        self.log("step", self.current_epoch, on_epoch=True)
        # calc average epoch loss and log it
        epoch_end_compute_and_log_losses(self, "validation", [e["losses"] for e in step_outputs])
        # evaluate  and log it
        epoch_end_compute_and_log_metrics(self, "validation", self._validation_metrics)

    def test_epoch_end(self, step_outputs) -> None:
        # for the logs to be at each epoch, not each step
        self.log("step", self.current_epoch, on_epoch=True)
        # calc average epoch loss and log it
        epoch_end_compute_and_log_losses(self, "test", [e["losses"] for e in step_outputs])
        # evaluate  and log it
        epoch_end_compute_and_log_metrics(self, "test", self._test_metrics)

    # confiugration
    def configure_callbacks(self) -> Sequence[pl.Callback]:
        """train loop callbacks"""
        return self._callbacks

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """See pl.LightningModule.configure_optimizers return value for all options"""
        return self._optimizers_and_lr_schs

    def set_predictions_keys(self, keys: List[str]) -> None:
        """Define which keys to extract from batch_dict on prediction mode"""
        self._prediction_keys = keys



import os
from typing import Optional, List, Tuple
import getpass
from .shell_handler import ShellHandler
from os.path import abspath, join, dirname
from glob import glob
from collections import namedtuple

SCRIPT_RUNNER_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "script_runner")


def get_script_runner_path(postfix="_DETACHED"):
    return SCRIPT_RUNNER_PATH + postfix + ".sh"


RemoteCommand = namedtuple("RemoteCommand", "machine command")


class RemoteExecution:
    def __init__(
        self,
        conda_env: str,
        password: str,
        user: Optional[str] = None,
        allow_manual_password_typing=True,
        verbose=0,
    ):
        self._conda_env = conda_env

        if password is None:
            if allow_manual_password_typing:
                password = getpass.getpass(prompt="password:")
            else:
                raise Exception("No password provided (and allow_manual_password_typing is disabled)")
        self._password = password

        if user is None:
            user = getpass.getuser()
        self._user = user
        self._verbose = verbose

    def run_remote_machine_command(self, machine: str, cmd: str) -> Tuple[str, str]:
        """
        connects to [machine] with ssh and runs [cmd].
        For example:
        run_remote_machine_command('server12', 'nvidia-smi')
        """
        shell = ShellHandler(machine, self._user, self._password)
        shin, shout, sherr = shell.execute(cmd)
        if self._verbose > 0:
            print("ssh returned:")
            print("shout:")
            _ = [print(x) for x in shout]
            print("sherr:")
            _ = [print(x) for x in sherr]
            print("\n\n\n")
        return shout, sherr

    def run_remote_python_command(
        self,
        *python_args,
        machine: str = None,
        gpu: str = None,
        log_output: str = None,
        auto_add_repo_root_dirs_to_PYTHONPATH=False,
        verbose=0,
    ) -> Tuple[str, str]:
        """
        connects to [machine] with ssh and runs [cmd].
        :param gpu: str that will be used directly as CUDA_VISIBLE_DEVICES
        For example:
        run_remote_python_command(
            '/path/to/some/script.py',
            '12',
            4,
            'banana',
            #

            machine='server12',
            gpu='1,4',
            log_output='./my_log1.txt',
        )

        this will run the python 'script.py', passing args '12 4 banana' to it.
        It will run it on machine 'server12', enabling gpus 1 and 4,
        and logging stdout+stderr to 'my_log1.txt'
        """

        print(f"total of {len(python_args)} python args")
        for a in python_args:
            print(a)

        assert len(python_args) > 0
        python_args = " ".join(python_args)

        assert isinstance(machine, str)
        assert isinstance(gpu, str)
        assert isinstance(log_output, str)

        if auto_add_repo_root_dirs_to_PYTHONPATH:
            all_repos_base_dir = abspath(join(dirname(abspath(__file__)), "../../"))
            repo_dirs = glob(all_repos_base_dir + "/*/")
            PYTHONPATH = ":".join(repo_dirs)
        else:
            PYTHONPATH = "/dummy"  # it is a positional arg, so we gotta give it something

        script_runner = get_script_runner_path()

        cmd = f"bash {script_runner} "
        cmd += f"{gpu} {self._conda_env} {PYTHONPATH} {log_output} "
        cmd += python_args

        print(cmd)
        # assert False #

        shell = ShellHandler(machine, self._user, self._password)
        shin, shout, sherr = shell.execute(cmd)
        if verbose > 0:
            print("ssh returned:")
            print("shout:")
            _ = [print(x) for x in shout]
            print("sherr:")
            _ = [print(x) for x in sherr]
            print("\n\n\n")
        return shout, sherr

    def run_multi_machines_commands(self, commands: List[RemoteCommand], verbose=0) -> None:
        """
        Execute a list of RemoteCommand-s on remote machines

        For example:
        cmds = [
            RemoteCommand(machine='server11', cmd='echo test123'),
            RemoteCommand(machine='server23', cmd='echo test456'),
        ]
        run_multi_machines_commands(cmds)
        """

        _ERR_TEXT = '"commands" is expected to be a list of RemoteCommand instances'

        if not isinstance(commands, list):
            raise Exception(_ERR_TEXT)
        if 0 == len(commands):
            raise Exception("commands is empty")
        if not isinstance(commands[0], RemoteCommand):
            raise Exception(_ERR_TEXT)

        for rem_cmd in commands:
            self.run_remote_machine_command(rem_cmd.machine, rem_cmd.command, verbose=verbose)



import os
import json
import pandas as pd
import numpy as np

CLINICAL_NAMES = [
    "SubjectId",
    "age",
    "bmi",
    "gender",
    "gender_num",
    "comorbidities",
    "smoking_history",
    "radiographic_size",
    "preop_egfr",
    "pathology_t_stage",
    "pathology_n_stage",
    "pathology_m_stage",
    "grade",
    "aua_risk_group",
    "task_1_label",
    "task_2_label",
]


def create_knight_clinical(original_file, processed_file=None):
    with open(original_file) as f:
        clinical_data = json.load(f)
    t_stage_count = np.zeros((5))
    aua_risk_count = np.zeros((5))
    df = pd.DataFrame(columns=CLINICAL_NAMES)
    for index, patient in enumerate(clinical_data):
        df.loc[index, "SubjectId"] = patient["case_id"]
        df.loc[index, "age"] = patient["age_at_nephrectomy"]
        df.loc[index, "bmi"] = patient["body_mass_index"]

        df.loc[index, "gender"] = patient["gender"]
        if patient["gender"] == "male":  # 0:'male'  1:'female','transgender_male_to_female'
            df.loc[index, "gender_num"] = 0
        else:
            df.loc[index, "gender_num"] = 1

        df.loc[index, "comorbidities"] = 0  # 0:no_comorbidities 1:comorbidities_exist
        for key, value in patient["comorbidities"].items():
            if value:
                df.loc[index, "comorbidities"] = 1

        df.loc[index, "smoking_history"] = patient["smoking_history"]
        if patient["smoking_history"] == "never_smoked":  # 0:'never_smoked' 1:'previous_smoker'  2:'current_smoker'
            df.loc[index, "smoking_history"] = 0
        elif patient["smoking_history"] == "previous_smoker":
            df.loc[index, "smoking_history"] = 1
        elif patient["smoking_history"] == "current_smoker":
            df.loc[index, "smoking_history"] = 2

        df.loc[index, "radiographic_size"] = patient["radiographic_size"]
        if patient["last_preop_egfr"]["value"] == ">=90":
            df.loc[index, "preop_egfr"] = 90
        else:
            df.loc[index, "preop_egfr"] = patient["last_preop_egfr"]["value"]

        df.loc[index, "pathology_t_stage"] = patient["pathology_t_stage"]
        df.loc[index, "pathology_n_stage"] = patient["pathology_n_stage"]
        df.loc[index, "pathology_m_stage"] = patient["pathology_m_stage"]
        df.loc[index, "aua_risk_group"] = patient["aua_risk_group"]

        # Task 1 labels:
        if patient["aua_risk_group"] in ["high_risk", "very_high_risk"]:  # 1:'3','4'  0:'0','1a','1b','2a','2b'
            df.loc[index, "task_1_label"] = 1  # CanAT
        else:
            df.loc[index, "task_1_label"] = 0  # NoAT

        # Task 2 labels:
        if patient["aua_risk_group"] == "benign":
            df.loc[index, "task_2_label"] = 0
        elif patient["aua_risk_group"] == "low_risk":
            df.loc[index, "task_2_label"] = 1
        elif patient["aua_risk_group"] == "intermediate_risk":
            df.loc[index, "task_2_label"] = 2
        elif patient["aua_risk_group"] == "high_risk":
            df.loc[index, "task_2_label"] = 3
        elif patient["aua_risk_group"] == "very_high_risk":
            df.loc[index, "task_2_label"] = 4
        else:
            ValueError("Wrong risk class")

        # former classification - deprecated
        # if patient['pathology_t_stage'] in ['3', '4']:    # 1:'3','4'  0:'0','1a','1b','2a','2b'
        #    df.loc[index, 'pathology_t_stage_classify'] = 1
        # else:
        #    df.loc[index, 'pathology_t_stage_classify'] = 0
        t_stage = int(patient["pathology_t_stage"][0])
        t_stage_count[t_stage] += 1
        aua_risk = int(df.loc[index, "task_2_label"])
        aua_risk_count[aua_risk] += 1
        df.loc[index, "grade"] = patient["tumor_isup_grade"]

    if processed_file is not None:
        # save csv file
        df.to_csv(processed_file, index=False)
        df = df.drop(["gender", "pathology_t_stage", "pathology_n_stage", "pathology_m_stage"], axis=1)
        df.to_csv(os.path.splitext(processed_file)[0] + "_numeric.csv", index=False)
    print(f"Pathology t-stage count summary: {t_stage_count}")
    print(f"AUA risk count summary: {aua_risk_count}")
    return df



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Dict, Tuple, Sequence, Optional


class ClassifierLinear(nn.Module):
    def __init__(
        self,
        in_ch: Sequence[int] = (256,),
        num_classes: float = 2,
        layers_description: Sequence[int] = (256,),
        dropout_rate: float = 0.1,
    ):

        super(ClassifierLinear, self).__init__()

        layer_list = []
        if layers_description is not None:
            layer_list.append(nn.Linear(in_ch, layers_description[0]))
            layer_list.append(nn.ReLU())
            if dropout_rate is not None and dropout_rate > 0:
                layer_list.append(nn.Dropout(p=dropout_rate))
            last_layer_size = layers_description[0]
            if len(layers_description) > 1:
                for curr_layer_size in layers_description[1:]:
                    layer_list.append(nn.Linear(last_layer_size, curr_layer_size))
                    layer_list.append(nn.ReLU())
                    if dropout_rate is not None and dropout_rate > 0:
                        layer_list.append(nn.Dropout(p=dropout_rate))
                    last_layer_size = curr_layer_size
        else:
            last_layer_size = in_ch

        layer_list.append(nn.Linear(last_layer_size, num_classes))
        self.classifier = nn.Sequential(*layer_list)

    def forward(self, x):
        x = self.classifier(x)
        return x


class LinearLayers(nn.Module):
    def __init__(
        self,
        in_ch: Sequence[int] = (256,),
        layers_description: Sequence[int] = (256,),
        dropout_rate: float = 0.1,
    ):

        super(LinearLayers, self).__init__()
        layer_list = []
        layer_list.append(nn.Linear(in_ch, layers_description[0]))
        layer_list.append(nn.ReLU())
        if dropout_rate is not None and dropout_rate > 0:
            layer_list.append(nn.Dropout(p=dropout_rate))
        last_layer_size = layers_description[0]
        if len(layers_description) > 1:
            for curr_layer_size in layers_description[1:]:
                layer_list.append(nn.Linear(last_layer_size, curr_layer_size))
                layer_list.append(nn.ReLU())
                if dropout_rate is not None and dropout_rate > 0:
                    layer_list.append(nn.Dropout(p=dropout_rate))
                last_layer_size = curr_layer_size

        self.classifier = nn.Sequential(*layer_list)

    def forward(self, x):
        x = self.classifier(x)
        return x


class Head1DClassifier(nn.Module):
    def __init__(
        self,
        head_name: str = "head_0",
        conv_inputs: Sequence[Tuple[str, int]] = None,
        num_classes: int = 2,
        post_concat_inputs: Optional[Sequence[Tuple[str, int]]] = None,
        post_concat_model: Optional[Sequence[int]] = None,
        layers_description: Sequence[int] = (256,),
        dropout_rate: float = 0.1,
        shared_classifier_head: Optional[torch.nn.Module] = None,
    ) -> None:
        """
        Classifier head 1d.

        Output of a forward pass:
        'model.logits.head_name' and 'outputs.head_name', both in shape [batch_size, num_classes]

        :param head_name:                   batch_dict key
        :param conv_inputs:                 List of feature map inputs - tuples of (batch_dict key, channel depth)
                                            If multiple inputs are used, they are concatenated on the channel axis
                for example:
                conv_inputs=(('model.backbone_features', 193),)
        :param num_classes:                 Number of output classes (per feature map location)
        :param post_concat_inputs:          Additional vector (one dimensional) inputs, concatenated just before the classifier module
        :param post_concat_model            Layers description for the post_concat_inputs module - sequence of hidden layers sizes
        :param layers_description:          Layers description for the classifier module - sequence of hidden layers sizes
        :param dropout_rate:                Dropout rate for classifier module layers
        :param shared_classifier_head:      Optional reference for external torch.nn.Module classifier
        """
        super().__init__()

        self.head_name = head_name
        assert conv_inputs is not None, "conv_inputs must be provided"
        self.conv_inputs = conv_inputs
        self.post_concat_inputs = post_concat_inputs
        self.post_concat_model = post_concat_model

        feature_depth = sum([conv_input[1] for conv_input in self.conv_inputs])

        if (post_concat_inputs is not None) and (self.post_concat_model is None):
            ## concat post_concat_input directly to conv_input,
            feature_depth += sum([post_concat_input[1] for post_concat_input in post_concat_inputs])

        elif self.post_concat_model is not None:
            # concat post_concat_input features from classifier_post_concat_model to conv_input
            features_depth_post_concat = sum([post_concat_input[1] for post_concat_input in post_concat_inputs])
            self.classifier_post_concat_model = LinearLayers(
                in_ch=features_depth_post_concat, layers_description=self.post_concat_model, dropout_rate=dropout_rate
            )

        if shared_classifier_head is not None:
            self.classifier_head_module = shared_classifier_head
        else:
            self.classifier_head_module = ClassifierLinear(
                in_ch=feature_depth,
                num_classes=num_classes,
                layers_description=layers_description,
                dropout_rate=dropout_rate,
            )

    def forward(self, batch_dict: Dict) -> Dict:

        conv_input = torch.cat([batch_dict[conv_input[0]] for conv_input in self.conv_inputs])

        res = conv_input

        if self.post_concat_inputs is not None:
            post_concat_input = torch.cat(
                [batch_dict[post_concat_input[0]] for post_concat_input in self.post_concat_inputs]
            )
            if self.post_concat_model is None:
                # concat post_concat_input directly to conv_input
                res = torch.cat([res, post_concat_input], dim=1)
            else:
                # concat post_concat_input features from classifier_post_concat_model to conv_input
                post_concat_input_feat = self.classifier_post_concat_model(post_concat_input)
                res = torch.cat([res, post_concat_input_feat], dim=1)

        logits = self.classifier_head_module(res)  # --> res.shape = [batch_size, 2, 1, 1]
        if len(logits.shape) > 2:
            logits = logits.squeeze(dim=3)  # --> res.shape = [batch_size, 2, 1]
            logits = logits.squeeze(dim=2)  # --> res.shape = [batch_size, 2]

        cls_preds = F.softmax(logits, dim=1)

        batch_dict["model.logits." + self.head_name] = logits
        batch_dict["model.output." + self.head_name] = cls_preds

        return batch_dict



import errno
from typing import Iterable, List, Dict, Optional, Tuple, Union, Any
import pickle
import bz2
import gzip
import socket
import os
import numpy as np
import time
import datetime

import pandas as pd
import shutil
import h5py
import hdf5plugin
from scipy.io import arff

from fuse.utils.misc.misc import Misc

###note - it is required that hdf5 support will be installed in a way that blosc is supported as well
# the recommended way is to install it in the following way:
# pip install h5py hdf5plugin
# conda install -y pytables
# and when using, import tables  before everything else


def save_pickle(obj: Any, output_filename: str, compress: bool = False, verbose: int = 0) -> None:
    """
    Pickles object to a file, optionally compressing it.
    Note - when compress=True, gzip will be used, and a ".gz" extension will be adde if not already present.
    returns: the output_filename, potentially with the added extension.
    """
    if compress and not output_filename.endswith(".gz"):
        output_filename += ".gz"
    use_open = gzip.open if (compress or output_filename.endswith(".gz")) else open
    with use_open(output_filename, "wb") as f:
        pickle.dump(obj, f)
    if verbose > 0:
        print("saved pickle: ", output_filename)
    return output_filename


def load_pickle(filename: str) -> Any:
    """
    Loads content from a pickle file
    """
    use_open = open
    if filename.endswith(".gz"):
        use_open = gzip.open
    elif filename.endswith(".pbz2"):
        use_open = bz2.open
    with use_open(filename, "rb") as f:
        return pickle.load(f)


def save_pickle_safe(obj: Any, output_filename: str, compress: bool = False, verbose: bool = 0) -> None:
    """
    a multi-threading/multi-process safe version of save_pickle()
    """

    scrambed_filename = get_randomized_postfix_name(output_filename)

    use_open = gzip.open if compress else open
    if compress and not output_filename.endswith(".gz"):
        output_filename += ".gz"
        scrambed_filename += ".gz"

    with use_open(scrambed_filename, "wb") as f:
        pickle.dump(obj, f)

    os.rename(scrambed_filename, output_filename)
    if verbose > 0:
        print("saved pickle: ", output_filename)
    return output_filename


def save_text_file_safe(filename: str, str_content: str) -> None:
    """
    Saves str_content into a created text file.
    This function is multi-threading and multi-processing safe.
    """

    scrambed_filename = get_randomized_postfix_name(filename)
    save_text_file(scrambed_filename, str_content)
    os.rename(scrambed_filename, filename)
    return filename


def save_text_file(file_path: str, content: str = "") -> None:
    """
    Saves str_content into a created text file.
    """
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, "w") as f:
        f.write(content)
        f.close()


def read_simple_float_file(file_path: str) -> float:
    """
    Reads a float string from a file. File is expected to contain only the float number text.
    """
    with open(file_path, "r") as fh:
        # printl('succesfully opened session num file.')
        data = fh.read()
        float_value = float(data)
        return float_value


def read_simple_int_file(file_path: str) -> int:
    """
    Reads an int string from a file. File is expected to contain only the int number text.
    """
    with open(file_path, "r") as fh:
        # printl('succesfully opened session num file.')
        data = fh.read()
        int_value = int(data)
        return int_value


def read_text_file(file_path: str) -> str:
    """
    A simple helper function which loads text from a file
    """
    with open(file_path, "r") as fh:
        # printl('succesfully opened session num file.')
        data = fh.read()
        return data


G_host_name = socket.gethostname()


def get_randomized_postfix_name(filename: str, **additional_rand_state) -> str:
    """
    Returns a filename post-fixed with a random hash
    This is mostly used to create a multi-processing safe save operation.
    For example:
    ...
    desired_name = '/path/to/some/dir/file100.blah'
    scrambed_name = get_randomized_postfix_name(desired_name)
    np.save(scrambled_name, ...)
    os.rename(scrambled_name, desired_name)

    @param filename:
    @return:
    """

    import numpy as np
    import hashlib

    # out_dir = os.path.dirname(filename)
    str_for_hash = filename + str(additional_rand_state)
    str_for_hash += G_host_name + "@" + str(os.getpid()) + str(np.random.random())
    hash_val = hashlib.md5(str_for_hash.encode("utf-8")).hexdigest()

    # scrambed_filename = os.path.join(out_dir, hash_val+'.scrambed')
    scrambed_filename = filename + hash_val + ".scramlbed"
    return scrambed_filename


def read_single_str_line_file(file_path: str) -> str:
    with open(file_path, "r") as fh:
        data = fh.read()
        if data[-1] == "\n":
            data = data[:-1]
        return data


def create_simple_timestamp_file(file_path: str) -> None:
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, "w") as f:
        curr_time = time.time()
        time_stamp_str = datetime.datetime.fromtimestamp(curr_time).strftime("%Y-%m-%d %H:%M:%S")
        f.write(time_stamp_str)
        f.close()


def save_hdf5_safe(filename: str, use_blosc: bool = True, **kwarrays):
    """
    multi-threading and multi-processing safe saving content to hdf5 file
    args:
        use_blosc: uses the blosc compression algorithm
    """
    # first validate the request
    for k, d in kwarrays.items():
        if not isinstance(k, str):
            raise Exception(f"only str keys are supported, instead got {type(k)}")
        if not isinstance(d, np.ndarray):
            raise Exception(f"only np.ndarray data is supported, instead got {type(d)}")
    scrambed_filename = get_randomized_postfix_name(filename)
    # import ipdb;ipdb.set_trace()
    with h5py.File(scrambed_filename, "w") as h5f:
        for k, d in kwarrays.items():
            _use_kwargs = {}
            if use_blosc:
                _use_kwargs = hdf5plugin.Blosc()
            h5f.create_dataset(k, data=d, **_use_kwargs)

    os.rename(scrambed_filename, filename)  # '.' + saved_tensors_format)

    return filename


# TODO: CONSIDER supporting slicing more "organically" - for example, changing this into a class instance,
#      which supports something like: x = load_hdf5('blah'); x['a.b'][:2,10:20:3,...]


def load_hdf5(
    filename: str,
    custom_extract: Optional[Dict[str, Union[int, List, Tuple]]] = None,
) -> dict:
    """

    :param filename:
    :param custom_extraction - OPTIONAL:
        optional arg, by default all keys and data is extracted.

        If you want to extract only part of the stored arrays, and optionally sub-parts of individual arrays as well,
        pass a dict to custom_extract in the following format:
            key: hdf5 dataset name
            data: "index description"

        each "index description" can be:
            if None: returns the entire array data
            if an int: returns data[index, ...]

            if a list (or a tuple) of slice/Ellipsis elements to get a subset, for example:
            index=[slice(10,20,2), slice(None,20), Ellipsis] will extract data[10:20:2, :20, ...]
            usually resulting in much faster load time if it's a small part of the data
            (dependening also on alignment and data storage order)


        Usage example:

        load_hdf5('some_file.hdf5',
            {
                'a.b' : None, #will take all of it
                'a.z.c': [slice(1,10,2), slice(None,20), Ellipsis], #will extract data[1:10:2, :20, ...] from it
            }

        Note: if custom_extraction is provided (not None), then only data from keys found in it will be returned
        In the example above, let's say that there was also a dataset named 'z.z.banana' in the file, it will *not* be returned.
        )
    :return:
    """
    # import ipdb;ipdb.set_trace()

    ans = {}
    h5f = h5py.File(filename, "r")
    for k in h5f.keys():
        if custom_extract is not None:
            if k not in custom_extract:
                continue
            else:
                index = custom_extract[k]
        else:
            index = None

        dset = h5f[k]

        if index is None:
            np_arr = dset[:]
        elif isinstance(index, int):
            np_arr = dset[index, ...]
        else:
            assert isinstance(index, (list, tuple))
            assert len(index) > 0
            assert isinstance(
                index[0], (slice, type(Ellipsis))
            )  # checking just the first, but it should be all of them
            np_arr = dset[tuple(index)]
        ans[k] = np_arr

    return ans


def save_dataframe(df: pd.DataFrame, filename: str, **kwargs) -> None:
    """
    Save dataframe into a file. The file format inferred from filename suffix
    Supported types: "csv", "hd5", "hdf5", "pickle", "pkl", "gz", "xslx", "md"
    :param filename: path to the output file
    """
    file_type = filename.split(".")[-1]

    assert file_type in [
        "csv",
        "hd5",
        "hdf5",
        "hdf",
        "pickle",
        "pkl",
        "gz",
        "xslx",
        "md",
    ], f"file type {file_type} not supported"
    if file_type in ["pickle", "pkl", "gz"]:
        df.to_pickle(filename, **kwargs)
    elif file_type == "csv":
        df.to_csv(filename, **kwargs)
    elif file_type in ["hd5", "hdf5", "hdf"]:
        df.to_hdf(filename, **kwargs)
    elif file_type == "xslx":
        df.to_excel(filename, **kwargs)
    elif file_type == "md":
        df.to_markdown(filename, **kwargs)


def read_dataframe(filename: str) -> pd.DataFrame:
    """
    Read dataframe from a file. The file format inferred from filename suffix
    Supported types: "csv", "hd5", "hdf5", "pickle", "pkl", "gz", "xslx"
    :param filename: path to the output file
    """
    file_type = filename.split(".")[-1]

    assert file_type in [
        "csv",
        "hd5",
        "hdf5",
        "hdf",
        "pickle",
        "pkl",
        "gz",
        "xslx",
    ], f"file type {file_type} not supported"
    if file_type in ["pickle", "pkl", "gz"]:
        df = pd.read_pickle(filename)
    elif file_type == "csv":
        df = pd.read_csv(filename)
    elif file_type in ["hd5", "hdf5", "hdf"]:
        df = pd.read_hdf(filename)
    elif file_type == "xslx":
        df = pd.read_excel(filename)
    elif file_type == "arff":
        data = arff.loadarff(filename)
        df = pd.DataFrame(data[0])

    return df


def delete_directory_tree(del_dir: str) -> bool:
    """
    Will recursively delete del_dir
    :returns: True if was able to delete this dir, False if dir was not found
    Note: if the directory was found but could not delete it, an exception will be raised.
    """
    if not os.path.isdir(del_dir):
        return False
    try:
        shutil.rmtree(del_dir)
    except OSError as e:
        print(f"Error while trying to delete cache: {e.filename} - {e.strerror}.")
        raise

    return True


def create_dir(dir_path: str, error_if_exist: bool = False) -> None:
    """
    Create dir
    :param dir_path: path to dir. either relative or full
    :param error_if_exist: if false will ignore already exist error
    :return: None
    """
    if not os.path.isdir(dir_path):
        try:
            os.makedirs(dir_path)
        except OSError as e:
            if e.errno != errno.EEXIST or error_if_exist:
                raise


def remove_dir_content(dir_path: str, ignore_files: Iterable[str] = tuple(), force_reset: bool = False) -> None:
    """
    Remove the content of dir_path ignoring the files under ignore_files.
    If force_reset is False, prompts the user for approval before the deletion.

    :param dir_path: path to dir. either relative or full
    :param ignore_files: list of files to ignore (don't delete them)
    :param force_reset: when False (default), asks the user for approval when deleting content.
        Else, delete without prompting.
    :return: None
    """
    # if no content - do nothing
    files = os.listdir(dir_path)
    files = [file for file in files if file not in ignore_files]
    num_files = len(files)
    if num_files > 0:
        # prompt the user for approval
        force_remove = (
            True
            if force_reset
            else Misc.query_yes_no(question=f"Folder {dir_path} contains {num_files} files. Delete anyway?")
        )
        # iterate over all the content and delete it
        failure = False
        if force_remove:
            for filename in files:
                file_path = os.path.join(dir_path, filename)
                try:
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.unlink(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
                except Exception as e:
                    print("Failed to delete %s. Reason: %s" % (file_path, e))
                    failure = True

        # in case remove wasn't approved or failed to delete some of the files
        if not force_remove or failure:
            msg = f"Folder {dir_path} is already used, please remove it and rerun the program"
            print(msg)
            raise Exception(msg)


def create_or_reset_dir(dir_path: str, ignore_files: Iterable[str] = tuple(), force_reset: bool = False) -> None:
    """
    Create dir or reset it if already exists
    :param dir_path: path to dir. either relative or full
    :param ignore_files: list of files to ignore (don't delete them)
    :return: None
    """
    create_dir(dir_path)
    remove_dir_content(dir_path, ignore_files, force_reset)



from typing import Optional, Union

from fuse.eval.metrics.libs.calibration import Calibration
from .metrics_classification_common import MetricMultiClassDefault

from functools import partial


class MetricReliabilityDiagram(MetricMultiClassDefault):
    """
    Compute and save a plot of the Reliability Diagram, a plot depicting a model's
    accuracy vs. confidence scores. For a well calibrated model, the result should be the identity function.
    See: https://arxiv.org/pdf/1706.04599.pdf
    """

    def __init__(
        self,
        pred: str,
        target: str,
        num_bins: Optional[int] = 10,
        num_quantiles: Optional[int] = None,
        output_filename: Optional[str] = None,
        **kwargs
    ) -> None:
        """
        :param pred: key name for the model prediction scores
        :param target: key name for the ground truth target values
        :param num_bins: number of equal width bins in the diagram's x-axis.
        :param num_quantiles: alternatively, number of equal number of samples quantiles in the diagram's x-axis.
        :param output_filename: output filename to save the figure. if None, will just save the arguments to create the figure in results dictionary.
        """

        reliability_diagram = partial(
            Calibration.reliability_diagram,
            num_bins=num_bins,
            num_quantiles=num_quantiles,
            output_filename=output_filename,
        )
        super().__init__(pred=pred, target=target, metric_func=reliability_diagram, **kwargs)


class MetricECE(MetricMultiClassDefault):
    """
    Compute the Expected Calibration Error, a measure for a model output
    probability score miscalibration (deviation from true probability for correctness).
    See: https://arxiv.org/pdf/1706.04599.pdf
    """

    def __init__(
        self, pred: str, target: str, num_bins: Optional[int] = 10, num_quantiles: Optional[int] = None, **kwargs
    ):
        """
        :param pred: key name for the model prediction scores
        :param target: key name for the ground truth target values
        :param class_names: class names. required for multi-class classifiers
        """
        ece = partial(Calibration.ece, num_bins=num_bins, num_quantiles=num_quantiles)
        super().__init__(pred=pred, target=target, metric_func=ece, **kwargs)


class MetricFindTemperature(MetricMultiClassDefault):
    """
    Find the optimal "temperature" for calibrating the prediction logits
    to represent calibrated probabilities.
    See: https://arxiv.org/pdf/1706.04599.pdf
    """

    def __init__(self, pred: str, target: str, **kwargs):
        super().__init__(pred=pred, target=target, metric_func=Calibration.find_temperature, **kwargs)
        """
        :param pred: key name for the model prediction scores
        :param target: key name for the ground truth target values
        """


class MetricApplyTemperature(MetricMultiClassDefault):
    """
    Apply temperature for calibrating the prediction logits
    to represent calibrated probabilities.
    See: https://arxiv.org/pdf/1706.04599.pdf
    """

    def __init__(self, pred: str, temperature: Union[float, str, None] = None, **kwargs):
        super().__init__(
            pred=pred, target=None, temperature=temperature, metric_func=Calibration.apply_temperature, **kwargs
        )
        """
        :param pred: key name for the model prediction scores
        :param target: key name for the ground truth target values
        """



import unittest

from fuse.utils.cpu_profiling import Profiler


class TestTimer(unittest.TestCase):
    def test_profiler_1(self):
        """ """
        from itertools import combinations

        def dummy_slow():
            N = 140  # 1000
            K = 2
            for _ in range(100):
                all_combs = combinations(range(N), K)
                all_combs = list(all_combs)
            return all_combs

        def dummy_quick():
            a = 1000
            a += 1
            return a

        def get_combs():
            for _ in range(10):
                x = dummy_slow()
                y = dummy_quick()

            return x, y

        with Profiler("banana"):
            get_combs()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import os
import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# list of requirements
requirements = []
with open(os.path.join(HERE, "requirements.txt"), "r") as fh:
    for line in fh:
        if not line.startswith("#"):
            requirements.append(line.strip())

setup(
    name="fuse-med-ml-examples",
    description="examples package for 'https://github.com/BiomedSciAI/fuse-med-ml/'",
    long_description="examples package for 'https://github.com/BiomedSciAI/fuse-med-ml/'",
    long_description_content_type="text/markdown",
    url="https://github.com/BiomedSciAI/fuse-med-ml/",
    author="IBM Research - Machine Learning for Healthcare and Life Sciences",
    author_email="moshiko.raboh@ibm.com",
    packages=find_packages(),
    license="Apache License 2.0",
    install_requires=requirements,
)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import List, Optional, Union
import unittest

import pandas as pds
import numpy as np
import torch
from torch.utils.data.dataloader import DataLoader

from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.utils.collates import CollateDefault
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data.ops.op_base import OpBase
from fuse.data import get_sample_id


class OpCustomCollateDefTest(OpBase):
    def __call__(self, sample_dict: dict, **kwargs) -> Union[None, dict, List[dict]]:
        if get_sample_id(sample_dict) == "a":
            sample_dict["data.partial"] = 1
        return sample_dict


class TestCollate(unittest.TestCase):
    def test_collate_default(self):
        # datainfo
        data = {
            "sample_id": ["a", "b", "c", "d", "e"],
            "data.values": [7, 4, 9, 2, 4],
            "data.nps": [np.array(4), np.array(2), np.array(5), np.array(1), np.array(4)],
            "data.torch": [torch.tensor(7), torch.tensor(4), torch.tensor(9), torch.tensor(2), torch.tensor(4)],
            "data.not_important": [12] * 5,
        }
        df = pds.DataFrame(data)

        # create simple pipeline
        op_df = OpReadDataframe(df)
        op_partial = OpCustomCollateDefTest()
        pipeline = PipelineDefault("test", [(op_df, {}), (op_partial, {})])

        # create dataset
        dataset = DatasetDefault(data["sample_id"], dynamic_pipeline=pipeline)
        dataset.create()

        # Use the collate function
        dl = DataLoader(
            dataset, 3, collate_fn=CollateDefault(skip_keys=["data.not_important"], raise_error_key_missing=False)
        )
        batch = next(iter(dl))

        # verify
        self.assertTrue("data.sample_id" in batch)
        self.assertListEqual(batch["data.sample_id"], ["a", "b", "c"])
        self.assertTrue((batch["data.values"] == torch.tensor([7, 4, 9])).all())
        self.assertTrue("data.nps" in batch)
        self.assertTrue((batch["data.nps"] == torch.stack([torch.tensor(4), torch.tensor(2), torch.tensor(5)])).all())
        self.assertTrue("data.torch" in batch)
        self.assertTrue((batch["data.torch"] == torch.stack([torch.tensor(7), torch.tensor(4), torch.tensor(9)])).all())
        self.assertTrue("data.partial" in batch)
        self.assertListEqual(batch["data.partial"], [1, None, None])
        self.assertFalse("data.not_important" in batch)

    def test_pad_all_tensors_to_same_size(self):
        a = torch.zeros((1, 1, 3))
        b = torch.ones((1, 2, 1))
        values = CollateDefault.pad_all_tensors_to_same_size([a, b])

        self.assertTrue((np.array(values.shape[1:]) == np.maximum(a.shape, b.shape)).all())
        self.assertTrue((values[1][:, :, :1] == b).all())
        self.assertTrue(values[1].sum() == b.sum())

    def test_pad_all_tensors_to_same_size_bs_1(self):
        a = torch.ones((1, 2, 1))
        values = CollateDefault.pad_all_tensors_to_same_size([a])
        self.assertTrue((values[0] == a).all())

    def test_pad_all_tensors_to_same_size_bs_3(self):
        a = torch.ones((1, 2, 3))
        b = torch.ones((3, 2, 1))
        c = torch.ones((1, 3, 2))
        values = CollateDefault.pad_all_tensors_to_same_size([a, b, c])
        self.assertListEqual(list(values.shape), [3, 3, 3, 3])


if __name__ == "__main__":
    unittest.main()



"""

(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021

"""

from collections import OrderedDict
import os
import sys
import copy
from fuse.eval.metrics.classification.metrics_thresholding_common import MetricApplyThresholds

from fuse.utils.utils_debug import FuseDebug
from fuse.utils.gpu import choose_and_enable_multiple_gpus

import logging

import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data.dataloader import DataLoader

from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils import NDict
from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.split import dataset_balanced_division_to_folds
from fuse.dl.models import ModelMultiHead
from fuse.dl.models.heads.head_global_pooling_classifier import HeadGlobalPoolingClassifier
from fuse.dl.losses.loss_default import LossDefault

from fuse.eval.metrics.classification.metrics_classification_common import MetricAUCROC, MetricAccuracy
from fuseimg.datasets.cmmd import CMMD
from fuse.dl.models.backbones.backbone_inception_resnet_v2 import BackboneInceptionResnetV2
from fuse.dl.lightning.pl_funcs import convert_predictions_to_dataframe
from pytorch_lightning import Trainer
from fuse.dl.lightning.pl_module import LightningModuleDefault
from fuse.utils.file_io.file_io import create_dir, load_pickle, save_dataframe
from fuse.eval.evaluator import EvaluatorDefault
import torch
import hydra
from omegaconf import DictConfig, OmegaConf

assert (
    "CMMD_DATA_PATH" in os.environ
), "Expecting environment variable CMMD_DATA_PATH to be set. Follow the instruction in example README file to download and set the path to the data"
##########################################
# Debug modes
##########################################
mode = "default"  # Options: 'default', 'debug'. See details in FuseDebug
debug = FuseDebug(mode)


def create_model(train: NDict, paths: NDict) -> torch.nn.Module:
    """
    creates the model
    See HeadGlobalPoolingClassifier for details
    """
    if train["target"] == "classification":
        num_classes = 2
        gt_label = "data.gt.classification"
        skip_keys = ["data.gt.subtype"]
        class_names = ["Benign", "Malignant"]
    elif train["target"] == "subtype":
        num_classes = 4
        gt_label = "data.gt.subtype"
        skip_keys = ["data.gt.classification"]
        class_names = ["Luminal A", "Luminal B", "HER2-enriched", "triple negative"]
    else:
        raise ("unsuported target!!")
    model = ModelMultiHead(
        conv_inputs=(("data.input.img", 1),),
        backbone=BackboneInceptionResnetV2(input_channels_num=1),
        heads=[
            HeadGlobalPoolingClassifier(
                head_name="head_0",
                dropout_rate=0.5,
                conv_inputs=[("model.backbone_features", 384)],
                layers_description=(256,),
                num_classes=num_classes,
                pooling="avg",
            ),
        ],
    )
    # create lightining trainer.
    pl_trainer = Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train["trainer"]["num_epochs"],
        accelerator=train["trainer"]["accelerator"],
        devices=train["trainer"]["devices"],
        num_sanity_val_steps=-1,
        auto_select_gpus=True,
    )
    return model, pl_trainer, num_classes, gt_label, skip_keys, class_names


#################################
# Train Template
#################################
def run_train(paths: NDict, train: NDict) -> torch.nn.Module:
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")

    # Download data
    # ==============================================================================
    # Model
    # ==============================================================================
    lgr.info("Model:", {"attrs": "bold"})

    model, pl_trainer, num_classes, gt_label, skip_keys, class_names = create_model(train, paths)
    lgr.info("Model: Done", {"attrs": "bold"})

    lgr.info("\nFuse Train", {"attrs": ["bold", "underline"]})

    lgr.info(f'model_dir={paths["model_dir"]}', {"color": "magenta"})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {"color": "magenta"})

    #### Train Data
    # split to folds randomly - temp
    dataset_all = CMMD.dataset(
        paths["data_dir"],
        paths["data_misc_dir"],
        train["target"],
        paths["cache_dir"],
        reset_cache=False,
        num_workers=train["num_workers"],
        train=True,
    )
    folds = dataset_balanced_division_to_folds(
        dataset=dataset_all,
        output_split_filename=os.path.join(paths["data_misc_dir"], paths["data_split_filename"]),
        id="data.patientID",
        keys_to_balance=[gt_label],
        nfolds=train["num_folds"],
        workers=train["num_workers"],
    )

    train_sample_ids = []
    for fold in train["train_folds"]:
        train_sample_ids += folds[fold]
    validation_sample_ids = []
    for fold in train["validation_folds"]:
        validation_sample_ids += folds[fold]

    train_dataset = CMMD.dataset(
        paths["data_dir"],
        paths["data_misc_dir"],
        train["target"],
        paths["cache_dir"],
        reset_cache=False,
        num_workers=train["num_workers"],
        sample_ids=train_sample_ids,
        train=True,
    )

    validation_dataset = CMMD.dataset(
        paths["data_dir"],
        paths["data_misc_dir"],
        train["target"],
        paths["cache_dir"],
        reset_cache=False,
        num_workers=train["num_workers"],
        sample_ids=validation_sample_ids,
    )

    ## Create sampler
    lgr.info("- Create sampler:")
    sampler = BatchSamplerDefault(
        dataset=train_dataset,
        balanced_class_name=gt_label,
        num_balanced_classes=num_classes,
        batch_size=train["batch_size"],
        mode="approx",
        workers=train["num_workers"],
        balanced_class_weights=None,
    )

    lgr.info("- Create sampler: Done")

    ## Create dataloader
    train_dataloader = DataLoader(
        dataset=train_dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=sampler,
        collate_fn=CollateDefault(skip_keys=skip_keys),
        num_workers=train["num_workers"],
    )
    lgr.info("Train Data: Done", {"attrs": "bold"})

    #### Validation data
    lgr.info("Validation Data:", {"attrs": "bold"})

    ## Create dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=train["batch_size"],
        num_workers=train["num_workers"],
        collate_fn=CollateDefault(skip_keys=skip_keys),
    )
    lgr.info("Validation Data: Done", {"attrs": "bold"})

    # ====================================================================================
    #  Loss
    # ====================================================================================
    losses = {
        "cls_loss": LossDefault(pred="model.logits.head_0", target=gt_label, callable=F.cross_entropy, weight=1.0)
    }

    # ====================================================================================
    # Metrics
    # ====================================================================================
    train_metrics = OrderedDict(
        [
            ("op", MetricApplyThresholds(pred="model.output.head_0")),  # will apply argmax
            ("auc", MetricAUCROC(pred="model.output.head_0", target=gt_label, class_names=class_names)),
            ("accuracy", MetricAccuracy(pred="results:metrics.op.cls_pred", target=gt_label)),
        ]
    )

    validation_metrics = copy.deepcopy(train_metrics)  # use the same metrics in validation as well

    # either a dict with arguments to pass to ModelCheckpoint or list dicts for multiple ModelCheckpoint callbacks (to monitor and save checkpoints for more then one metric).
    best_epoch_source = dict(
        monitor="validation.metrics.auc.macro_avg",
        mode="max",
    )

    # =====================================================================================
    #  Manager - Train
    #  Create a manager, training objects and run a training process.
    # =====================================================================================
    lgr.info("Train:", {"attrs": "bold"})

    # create optimizer
    optimizer = optim.Adam(model.parameters(), lr=train["learning_rate"], weight_decay=train["weight_decay"])
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)

    lr_sch_config = dict(scheduler=scheduler, monitor="validation.losses.total_loss")

    # optimizier and lr sch - see pl.LightningModule.configure_optimizers return value for all options
    optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    # create instance of PL module - FuseMedML generic version

    pl_module = LightningModuleDefault(
        model_dir=paths["model_dir"],
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=validation_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )

    # create lightining trainer.
    pl_trainer = Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train["trainer"]["num_epochs"],
        accelerator=train["trainer"]["accelerator"],
        devices=train["trainer"]["devices"],
        num_sanity_val_steps=-1,
        auto_select_gpus=True,
    )

    # train from scratch
    pl_trainer.fit(pl_module, train_dataloader, validation_dataloader, ckpt_path=train["trainer"]["ckpt_path"])
    lgr.info("Train: Done", {"attrs": "bold"})


######################################
# Inference Template
######################################
def run_infer(train: NDict, paths: NDict, infer: NDict):
    create_dir(paths["inference_dir"])
    #### Logger
    fuse_logger_start(output_path=paths["inference_dir"], console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Inference", {"attrs": ["bold", "underline"]})
    infer_file = os.path.join(paths["inference_dir"], infer["infer_filename"])
    checkpoint_file = os.path.join(paths["model_dir"], infer["checkpoint"])
    lgr.info(f"infer_filename={checkpoint_file}", {"color": "magenta"})

    lgr.info("Model:", {"attrs": "bold"})

    model, pl_trainer, num_classes, gt_label, skip_keys, class_names = create_model(train, paths)
    lgr.info("Model: Done", {"attrs": "bold"})
    ## Data
    folds = load_pickle(
        os.path.join(paths["data_misc_dir"], paths["data_split_filename"])
    )  # assume exists and created in train func

    infer_sample_ids = []
    for fold in infer["infer_folds"]:
        infer_sample_ids += folds[fold]

    test_dataset = CMMD.dataset(
        paths["data_dir"],
        paths["data_misc_dir"],
        infer["target"],
        paths["cache_dir"],
        sample_ids=infer_sample_ids,
        train=False,
    )
    ## Create dataloader
    infer_dataloader = DataLoader(
        dataset=test_dataset,
        shuffle=False,
        drop_last=False,
        collate_fn=CollateDefault(),
        num_workers=infer["num_workers"],
    )
    # load python lightning module
    pl_module = LightningModuleDefault.load_from_checkpoint(
        checkpoint_file, model_dir=paths["model_dir"], model=model, map_location="cpu", strict=True
    )
    # set the prediction keys to extract (the ones used be the evaluation function).
    pl_module.set_predictions_keys(
        ["model.output.head_0", "data.gt.classification"]
    )  # which keys to extract and dump into file
    lgr.info("Test Data: Done", {"attrs": "bold"})

    # create a trainer instance
    predictions = pl_trainer.predict(pl_module, infer_dataloader, return_predictions=True)

    # convert list of batch outputs into a dataframe
    infer_df = convert_predictions_to_dataframe(predictions)
    save_dataframe(infer_df, infer_file)


######################################
# Analyze Template
######################################
def run_eval(paths: NDict, infer: NDict):
    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Eval", {"attrs": ["bold", "underline"]})

    # metrics
    metrics = OrderedDict(
        [
            ("op", MetricApplyThresholds(pred="model.output.head_0")),  # will apply argmax
            ("auc", MetricAUCROC(pred="model.output.head_0", target="data.gt.classification")),
            ("accuracy", MetricAccuracy(pred="results:metrics.op.cls_pred", target="data.gt.classification")),
        ]
    )

    # create evaluator
    evaluator = EvaluatorDefault()

    # run
    results = evaluator.eval(
        ids=None,
        data=os.path.join(paths["inference_dir"], infer["infer_filename"]),
        metrics=metrics,
        output_dir=paths["eval_dir"],
    )

    return results


@hydra.main(config_path="conf", config_name="config")
def main(cfg: DictConfig) -> None:
    cfg = NDict(OmegaConf.to_object(cfg))
    print(cfg)
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    choose_and_enable_multiple_gpus(cfg["train.trainer.devices"], force_gpus=force_gpus)

    # Path to the stored dataset location
    # dataset should be download from https://wiki.cancerimagingarchive.net/pages/viewpage.action?pageId=70230508
    # download requires NBIA data retriever https://wiki.cancerimagingarchive.net/display/NBIA/Downloading+TCIA+Images
    # put on the following in the main folder  -
    # 1. CMMD_clinicaldata_revision.csv which is a converted version of CMMD_clinicaldata_revision.xlsx
    # 2. folder named CMMD which is the downloaded data folder

    # train
    if "train" in cfg["run.running_modes"]:
        run_train(cfg["paths"], cfg["train"])

    # infer
    if "infer" in cfg["run.running_modes"]:
        run_infer(cfg["train"], cfg["paths"], cfg["infer"])
    #
    # analyze
    if "eval" in cfg["run.running_modes"]:
        run_eval(cfg["paths"], cfg["infer"])


if __name__ == "__main__":
    sys.argv.append("hydra.run.dir=working_dir")
    main()



from typing import Sequence, Optional, Hashable
import numpy as np
import copy


class PerSampleData:
    """
    Basic class for sampling data at given sample ids.
    Used in order to maintain sample ids alignment with the source data when
    applying metrics and operations which operate per sample and combining them with
    other metrics in a pipeline.
    """

    def __init__(self, data: Sequence[np.ndarray], ids: Sequence[Hashable]):
        self._data = data
        self._ids = ids

    def __call__(self, ids: Optional[Sequence[Hashable]] = None) -> Sequence[np.ndarray]:
        if ids is None:
            return copy.copy(self._data)
        else:
            # convert required ids to permutation
            original_ids = self._ids
            required_ids = ids
            permutation = [original_ids.index(sample_id) for sample_id in required_ids]

            return [self._data[i] for i in permutation]



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Callable, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from fuse.dl.losses.loss_base import LossBase
from fuse.utils.ndict import NDict
import numpy as np


def make_one_hot(input, num_classes):
    """Convert class index tensor to one hot encoding tensor.
    Args:
         input: A tensor of shape [N, 1, *]
         num_classes: An int of number of class
    Returns:
        A tensor of shape [N, num_classes, *]
    """
    shape = np.array(input.shape)
    shape[1] = num_classes
    shape = tuple(shape)
    result = torch.zeros(shape, device=input.device)
    result = result.scatter_(1, input, 1)

    return result


class WeightedFocalLoss(nn.Module):
    "Non weighted version of Focal Loss"

    def __init__(self, alpha: float = 0.25, gamma: float = 2.0):
        """
        FL = -alpha(1-pt)^gamma * log(pt)
        :param alpha: hyperparameter weighted term
        :param gamma: hyperparameter
        """
        super(WeightedFocalLoss, self).__init__()
        self.alpha = torch.tensor([alpha, 1 - alpha])
        self.gamma = gamma

    def forward(self, inputs, targets):

        if targets.dim() < 4:
            targets = targets.unsqueeze(1)
        targets = make_one_hot(targets, inputs.shape[1])

        BCE_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")
        targets = targets.type(torch.long)
        at = self.alpha.gather(0, targets.data.view(-1))
        pt = torch.exp(-BCE_loss).data.view(-1)
        F_loss = at * (1 - pt) ** self.gamma * BCE_loss.data.view(-1)
        return F_loss.mean()


class FocalLoss(LossBase):
    def __init__(
        self,
        pred_name: str = None,
        target_name: str = None,
        weight: float = 1.0,
        alpha: float = 0.25,
        gamma: float = 2,
        filter_func: Optional[Callable] = None,
        resize_mode: str = "interpolate",
    ) -> None:
        """
        :param pred_name:               batch_dict key for predicted logits (e.g., class probabilities BEFORE softmax).
                                        Expected Tensor shape = [batch, num_classes, height, width]
        :param target_name:             batch_dict key for target (e.g., ground truth label). Expected Tensor shape = [batch, height, width]
        :param callable:                PyTorch loss function handle (e.g., torch.nn.functional.cross_entropy)
        :param alpha:                   hyperparameter for focal loss function
        :param gamma:                   hyperparameter for focal loss function
        :param weight:                  Weight multiplier for final loss value
        :param filter_func:             function that filters batch_dict/ The function gets ans input batch_dict and returns filtered batch_dict
        :param resize_mode:             Resize mode- either using a max pooling kernel(default), or using PyTorch interpolation ('interpolate'/'maxpool')
        """
        super().__init__(pred_name, target_name, weight)
        self.filter_func = filter_func
        self.resize_mode = resize_mode
        self.callable = WeightedFocalLoss(alpha=alpha, gamma=gamma)

    def __call__(self, batch_dict: NDict) -> torch.Tensor:
        # filter batch_dict if required
        if self.filter_func is not None:
            batch_dict = self.filter_func(batch_dict)

        preds = batch_dict[self.pred_name]  # preds shape [batch_size, num_classes, height, width]
        targets = batch_dict[self.target_name]  # targets shape [batch_size, height, width]

        batch_size, targets_height, targets_width = targets.shape
        batch_size, num_classes, preds_height, preds_width = preds.shape

        # Resize targets (if needed) to match predicted activation map
        # ============================================================
        targets = targets.reshape([batch_size, 1, targets_height, targets_width])

        if (targets_height > preds_height) or (targets_width > preds_width):
            if targets.dtype != torch.float32:
                targets = targets.type(torch.float32).to(targets.device)
            if self.resize_mode == "maxpool":
                block_height = int(targets_height / preds_height)
                block_width = int(targets_width / preds_width)
                residual_h = int((targets_height - (block_height * preds_height)) / 2)
                residual_w = int((targets_width - (block_width * preds_width)) / 2)

                targets = torch.nn.functional.max_pool2d(
                    targets[:, :, residual_h : targets_height - residual_h, residual_w : targets_width - residual_w],
                    kernel_size=(block_height, block_width),
                )
            elif self.resize_mode == "interpolate":
                targets = torch.nn.functional.interpolate(targets, size=(preds_height, preds_width))
            else:
                raise Exception

        # Reshape targets and change its dtype to int64 for cross entropy input
        # =====================================================================
        targets = targets.reshape([batch_size, preds_height, preds_width])
        if targets.dtype != torch.int64:
            targets = targets.type(torch.int64).to(targets.device)

        return self.callable(preds, targets) * self.weight



from typing import List, Optional
from fuse.data.ops.op_base import OpBase
from fuse.utils.ndict import NDict
from fuse.utils.rand.param_sampler import Gaussian
from fuse.data.ops.ops_cast import Cast
from fuseimg.data.ops.color import OpClip
from torch import Tensor
import torch


class OpAugColor(OpBase):
    """
    Color augmentation for gray scale images of any dimensions, including addition, multiplication, gamma and contrast adjusting
    """

    def __init__(self, verify_arguments: bool = True):
        """
        :param verify_arguments: this op expects torch tensor of range [0, 1]. Set to False to disable verification
        """
        super().__init__()
        self._verify_arguments = verify_arguments

    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        add: Optional[float] = None,
        mul: Optional[float] = None,
        gamma: Optional[float] = None,
        contrast: Optional[float] = None,
        channels: Optional[List[int]] = None,
    ):
        """
        :param key: key to a image stored in sample_dict: torch tensor of range [0, 1] representing an image to ,
        :param add: value to add to each pixel
        :param mul: multiplication factor
        :param gamma: gamma factor
        :param contrast: contrast factor
        :param channels: Apply clipping just over the specified channels. If set to None will apply on all channels.
        """
        aug_input = sample_dict[key]

        # verify
        if self._verify_arguments:
            assert isinstance(aug_input, torch.Tensor), f"Error: OpAugColor expects torch Tensor, got {type(aug_input)}"
            assert (
                aug_input.min() >= 0.0 and aug_input.max() <= 1.0
            ), f"Error: OpAugColor expects tensor in range [0.0-1.0]. got [{aug_input.min()}-{aug_input.max()}]"

        aug_tensor = aug_input
        if channels is None:
            if add is not None:
                aug_tensor = self.aug_op_add_col(aug_tensor, add)
            if mul is not None:
                aug_tensor = self.aug_op_mul_col(aug_tensor, mul)
            if gamma is not None:
                aug_tensor = self.aug_op_gamma(aug_tensor, 1.0, gamma)
            if contrast is not None:
                aug_tensor = self.aug_op_contrast(aug_tensor, contrast)
        else:
            if add is not None:
                aug_tensor[channels] = self.aug_op_add_col(aug_tensor[channels], add)
            if mul is not None:
                aug_tensor[channels] = self.aug_op_mul_col(aug_tensor[channels], mul)
            if gamma is not None:
                aug_tensor[channels] = self.aug_op_gamma(aug_tensor[channels], 1.0, gamma)
            if contrast is not None:
                aug_tensor[channels] = self.aug_op_contrast(aug_tensor[channels], contrast)

        sample_dict[key] = aug_tensor
        return sample_dict

    @staticmethod
    def aug_op_add_col(aug_input: Tensor, add: float) -> Tensor:
        """
        Adding a values to all pixels
        :param aug_input: the tensor to augment
        :param add: the value to add to each pixel
        :return: the augmented tensor
        """
        aug_tensor = aug_input + add
        aug_tensor = OpClip.clip(aug_tensor, clip=(0.0, 1.0))
        return aug_tensor

    @staticmethod
    def aug_op_mul_col(aug_input: Tensor, mul: float) -> Tensor:
        """
        multiply each pixel
        :param aug_input: the tensor to augment
        :param mul: the multiplication factor
        :return: the augmented tensor
        """
        input_tensor = aug_input * mul
        input_tensor = OpClip.clip(input_tensor, clip=(0.0, 1.0))
        return input_tensor

    @staticmethod
    def aug_op_gamma(aug_input: Tensor, gain: float, gamma: float) -> Tensor:
        """
        Gamma augmentation
        :param aug_input: the tensor to augment
        :param gain: gain factor
        :param gamma: gamma factor
        :return: None
        """
        input_tensor = (aug_input**gamma) * gain
        input_tensor = OpClip.clip(input_tensor, clip=(0.0, 1.0))
        return input_tensor

    @staticmethod
    def aug_op_contrast(aug_input: Tensor, factor: float) -> Tensor:
        """
        Adjust contrast (notice - calculated across the entire input tensor, even if it's 3d)
        :param aug_input:the tensor to augment
        :param factor: contrast factor.   1.0 is neutral
        :return: the augmented tensor
        """
        calculated_mean = aug_input.mean()
        input_tensor = ((aug_input - calculated_mean) * factor) + calculated_mean
        input_tensor = OpClip.clip(input_tensor, clip=(0.0, 1.0))
        return input_tensor


class OpAugGaussian(OpBase):
    """
    Add gaussian noise to numpy array or torch tensor of any dimensions
    """

    def __call__(
        self, sample_dict: NDict, key: str, mean: float = 0.0, std: float = 0.03, channels: Optional[List[int]] = None
    ) -> Tensor:
        """
        :param key: key to a tensor or numpy array stored in sample_dict: any dimension and any range
        :param mean: mean gaussian distribution
        :param std:  std gaussian distribution
        :param channels: Apply just over the specified channels. If set to None will apply on all channels.
        """
        aug_input = sample_dict[key]

        aug_tensor = aug_input
        if channels is None:
            rand_patch = Gaussian(aug_tensor.shape, mean, std).sample()
            rand_patch = Cast.like(rand_patch, aug_tensor)
            aug_tensor = aug_tensor + rand_patch

        else:
            rand_patch = Gaussian(aug_tensor[channels].shape, mean, std).sample()
            rand_patch = Cast.like(rand_patch, aug_tensor)
            aug_tensor[channels] = aug_tensor[channels] + rand_patch

        sample_dict[key] = aug_tensor
        return sample_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest
import os
import tempfile
import shutil
from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess

from fuse_examples.imaging.classification.isic.isic_runner import (
    TRAIN_COMMON_PARAMS,
    INFER_COMMON_PARAMS,
    EVAL_COMMON_PARAMS,
    run_train,
    run_infer,
    run_eval,
    PATHS,
)
from fuse_examples.imaging.classification.isic.golden_members import FULL_GOLDEN_MEMBERS

import fuse.utils.gpu as GPU
from fuse.utils.rand.seed import Seed


def run_isic(root: str) -> None:
    model_dir = os.path.join(root, "model_dir")
    paths = {
        "model_dir": model_dir,
        "data_dir": os.environ["ISIC19_DATA_PATH"]
        if "ISIC19_DATA_PATH" in os.environ
        else os.path.join(root, "isic/data_dir"),
        "cache_dir": os.path.join(root, "cache_dir"),
        "inference_dir": os.path.join(model_dir, "infer_dir"),
        "eval_dir": os.path.join(model_dir, "eval_dir"),
        "data_split_filename": os.path.join(root, "split.pkl"),
    }

    train_common_params = TRAIN_COMMON_PARAMS
    train_common_params["trainer.num_epochs"] = 2
    train_common_params["samples_ids"] = FULL_GOLDEN_MEMBERS

    infer_common_params = INFER_COMMON_PARAMS
    eval_common_params = EVAL_COMMON_PARAMS

    # Must use GPU due a long running time
    GPU.choose_and_enable_multiple_gpus(1, use_cpu_if_fail=False)

    Seed.set_seed(0, False)  # previous test (in the pipeline) changed the deterministic behavior to True

    run_train(paths, train_common_params)
    run_infer(paths, infer_common_params)
    results = run_eval(paths, eval_common_params)

    assert "metrics.auc" in results


class ClassificationISICTestCase(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()

    def test_runner(self):
        run_in_subprocess(run_isic, self.root)

    def tearDown(self):
        # Delete temporary directories
        shutil.rmtree(self.root)


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021
"""

from copy import deepcopy
import os
from typing import Optional
from fuse.utils.misc.misc import get_pretty_dataframe

import pytorch_lightning as pl
from pytorch_lightning.callbacks.base import Callback
import torch
import pandas as pd


class ModelEpochSummary(Callback):
    """
        Model Checkpointing Display
        ===================

        Automatically display (print to screen and log to a file) best vs current epoch metircs and losses.

        Example:
    Stats for epoch: 9 (best tpoch is 7 for source validation.metrics.accuracy!)
    ------------------------------------------------------------------------------------------
    |                             | Best Epoch (7)              | Current Epoch (9)           |
    ------------------------------------------------------------------------------------------
    | train.losses.cls_loss       | 0.1546                      | 0.1146                      |
    ------------------------------------------------------------------------------------------
    | train.losses.total_loss     | 0.1546                      | 0.1146                      |
    ------------------------------------------------------------------------------------------
    | train.metrics.accuracy      | 0.9545                      | 0.9655                      |
    ------------------------------------------------------------------------------------------
    | validation.losses.cls_loss  | 0.0720                      | 0.0920                      |
    ------------------------------------------------------------------------------------------
    | validation.losses.total_loss| 0.0720                      | 0.0920                      |
    ------------------------------------------------------------------------------------------
    | validation.metrics.accuracy | 0.9776                      | 0.9756                      |
    ------------------------------------------------------------------------------------------

    """

    def __init__(
        self,
        dirpath: Optional[str] = None,
        filename: Optional[str] = None,
        monitor: Optional[str] = None,
        mode: str = "min",
    ):
        """
        :param dirpath: location of log file
        :param filename: specify a filename. If not set will use f"epoch_summary_{monitor}.txt"
        :param monitor: the metric name to track
        :param mode: either consider the "min" value to be best or the "max" value to be the best
        """
        super().__init__()
        self._monitor = monitor
        self._mode = mode
        self._dirpath = dirpath
        self._filename = filename if filename is not None else f"epoch_summary_{self._monitor}.txt"
        self._best_epoch_metrics = None
        self._best_epoch_index = None

    def print_epoch_summary_table(self, epoch_metircs: dict, epoch_source_index: int) -> None:
        """Generate, print and log the table"""

        def get_value_as_float_str(dict, key):
            val_as_str = "N/A"
            try:
                value = dict[key]
                val_as_str = "%.4f" % float(value)
            except:
                pass
            return val_as_str

        stats_table = pd.DataFrame(
            columns=["", f"Best Epoch ({self._best_epoch_index})", f"Current Epoch ({epoch_source_index})"]
        )
        idx = 0

        eval_keys = sorted(epoch_metircs.keys())
        for evaluator_name in eval_keys:
            current_str = get_value_as_float_str(epoch_metircs, evaluator_name)
            best_so_far_str = get_value_as_float_str(self._best_epoch_metrics, evaluator_name)

            stats_table.loc[idx] = [f"{evaluator_name}", best_so_far_str, current_str]
            idx += 1

        if self._best_epoch_index == epoch_source_index:
            epoch_title = (
                f"Stats for epoch: {epoch_source_index} (Currently the best epoch for source {self._monitor}!)"
            )
            print(epoch_title)
        else:
            epoch_title = f"Stats for epoch: {epoch_source_index} (Best epoch is {self._best_epoch_index} for source {self._monitor})"
            print(epoch_title)
        stats_as_str = get_pretty_dataframe(stats_table)
        print(stats_as_str)
        print(f"Model: {self._dirpath}")

        try:
            op = "w" if epoch_source_index == 0 else "a"  # we want to have all stats in one file per epoch
            with open(os.path.join(self._dirpath, self._filename), op) as sfile:
                sfile.write(epoch_title)
                sfile.write(stats_as_str)
        except Exception as error:
            print(f"Cannot write epoch summary to file {self._dirpath}/{self._filename}")
            print(error)

    def on_train_epoch_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        """Print summary at the end of the validation stage."""
        monitor_candidates = monitor_candidates = deepcopy(trainer.callback_metrics)
        current_epoch_metrics = monitor_candidates
        monitor_op = {"min": torch.lt, "max": torch.gt}[self._mode]
        if self._best_epoch_metrics is None or monitor_op(
            current_epoch_metrics[self._monitor], self._best_epoch_metrics[self._monitor]
        ):
            self._best_epoch_metrics = current_epoch_metrics
            self._best_epoch_index = trainer.current_epoch
        self.print_epoch_summary_table(current_epoch_metrics, trainer.current_epoch)
